from unittest import TestCase
from mock import Mock
from autosubmit.job.job_list import JobList
from bscearth.utils.date import parse_date, date2str
from bscearth.utils.config_parser import ConfigParserFactory
from autosubmit.job.job_list_persistence import JobListPersistenceDb
from autosubmit.job.job_common import Status
from random import randrange
from autosubmit.job.job import Job
from mock import patch

from autosubmit.job.job_grouping import JobGrouping

class TestJobGrouping(TestCase):

    def setUp(self):
        self.experiment_id = 'random-id'
        self.job_list = JobList(self.experiment_id, FakeBasicConfig, ConfigParserFactory(),
                                JobListPersistenceDb('.', '.'))
        self.parser_mock = Mock(spec='SafeConfigParser')

        # Basic workflow with SETUP, INI, SIM, POST, CLEAN
        self._createDummyJob('expid_SETUP', Status.READY)

        for date in ['19000101', '19000202']:
            for member in ['m1', 'm2']:
                job = self._createDummyJob('expid_' + date + '_' + member + '_' + 'INI', Status.WAITING, date, member)
                self.job_list.get_job_list().append(job)

        sections = ['SIM', 'POST', 'CLEAN']
        for section in sections:
            for date in ['19000101', '19000202']:
                for member in ['m1', 'm2']:
                    for chunk in [1, 2]:
                        job = self._createDummyJob('expid_' + date + '_' + member + '_' + str(chunk) + '_' + section, Status.WAITING, date, member, chunk)
                        self.job_list.get_job_list().append(job)

    def test_group_by_date(self):
        groups_dict = dict()

        groups_dict['status'] = {'19000101' : Status.WAITING, '19000202' : Status.WAITING}
        groups_dict['jobs'] = {
                                'expid_19000101_m1_INI' : ['19000101'], 'expid_19000101_m2_INI' : ['19000101'], 'expid_19000202_m1_INI' : ['19000202'], 'expid_19000202_m2_INI' : ['19000202'],

                               'expid_19000101_m1_1_SIM': ['19000101'], 'expid_19000101_m1_2_SIM': ['19000101'], 'expid_19000101_m2_1_SIM': ['19000101'], 'expid_19000101_m2_2_SIM': ['19000101'],
                               'expid_19000202_m1_1_SIM': ['19000202'], 'expid_19000202_m1_2_SIM': ['19000202'], 'expid_19000202_m2_1_SIM': ['19000202'], 'expid_19000202_m2_2_SIM': ['19000202'],

                               'expid_19000101_m1_1_POST': ['19000101'], 'expid_19000101_m1_2_POST': ['19000101'], 'expid_19000101_m2_1_POST': ['19000101'], 'expid_19000101_m2_2_POST': ['19000101'],
                               'expid_19000202_m1_1_POST': ['19000202'], 'expid_19000202_m1_2_POST': ['19000202'], 'expid_19000202_m2_1_POST': ['19000202'], 'expid_19000202_m2_2_POST': ['19000202'],

                               'expid_19000101_m1_1_CLEAN': ['19000101'], 'expid_19000101_m1_2_CLEAN': ['19000101'], 'expid_19000101_m2_1_CLEAN': ['19000101'],  'expid_19000101_m2_2_CLEAN': ['19000101'],
                               'expid_19000202_m1_1_CLEAN': ['19000202'], 'expid_19000202_m1_2_CLEAN': ['19000202'], 'expid_19000202_m2_1_CLEAN': ['19000202'], 'expid_19000202_m2_2_CLEAN': ['19000202']
                               }
        
        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.date is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('date', self.job_list.get_job_list(), self.job_list)
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_group_by_member(self):
        groups_dict = dict()

        groups_dict['status'] = {'19000101_m1': Status.WAITING, '19000101_m2': Status.WAITING, '19000202_m1': Status.WAITING, '19000202_m2' : Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI': ['19000101_m1'], 'expid_19000101_m2_INI': ['19000101_m2'], 'expid_19000202_m1_INI': ['19000202_m1'], 'expid_19000202_m2_INI': ['19000202_m2'],

            'expid_19000101_m1_1_SIM': ['19000101_m1'], 'expid_19000101_m1_2_SIM': ['19000101_m1'], 'expid_19000101_m2_1_SIM': ['19000101_m2'],
            'expid_19000101_m2_2_SIM': ['19000101_m2'],
            'expid_19000202_m1_1_SIM': ['19000202_m1'], 'expid_19000202_m1_2_SIM': ['19000202_m1'], 'expid_19000202_m2_1_SIM': ['19000202_m2'],
            'expid_19000202_m2_2_SIM': ['19000202_m2'],

            'expid_19000101_m1_1_POST': ['19000101_m1'], 'expid_19000101_m1_2_POST': ['19000101_m1'], 'expid_19000101_m2_1_POST': ['19000101_m2'],
            'expid_19000101_m2_2_POST': ['19000101_m2'],
            'expid_19000202_m1_1_POST': ['19000202_m1'], 'expid_19000202_m1_2_POST': ['19000202_m1'], 'expid_19000202_m2_1_POST': ['19000202_m2'],
            'expid_19000202_m2_2_POST': ['19000202_m2'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1'], 'expid_19000101_m2_1_CLEAN': ['19000101_m2'],
            'expid_19000101_m2_2_CLEAN': ['19000101_m2'],
            'expid_19000202_m1_1_CLEAN': ['19000202_m1'], 'expid_19000202_m1_2_CLEAN': ['19000202_m1'], 'expid_19000202_m2_1_CLEAN': ['19000202_m2'],
            'expid_19000202_m2_2_CLEAN': ['19000202_m2']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.member is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('member', self.job_list.get_job_list(), self.job_list)
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_group_by_chunk(self):
        groups_dict = dict()

        groups_dict['status'] = {'19000101_m1_1': Status.WAITING, '19000101_m1_2': Status.WAITING,
                                 '19000101_m2_1': Status.WAITING, '19000101_m2_2': Status.WAITING,
                                 '19000202_m1_1': Status.WAITING, '19000202_m1_2': Status.WAITING,
                                 '19000202_m2_1': Status.WAITING, '19000202_m2_2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_1_SIM': ['19000101_m1_1'], 'expid_19000101_m1_2_SIM': ['19000101_m1_2'], 'expid_19000101_m2_1_SIM': ['19000101_m2_1'],
            'expid_19000101_m2_2_SIM': ['19000101_m2_2'],
            'expid_19000202_m1_1_SIM': ['19000202_m1_1'], 'expid_19000202_m1_2_SIM': ['19000202_m1_2'], 'expid_19000202_m2_1_SIM': ['19000202_m2_1'],
            'expid_19000202_m2_2_SIM': ['19000202_m2_2'],

            'expid_19000101_m1_1_POST': ['19000101_m1_1'], 'expid_19000101_m1_2_POST': ['19000101_m1_2'], 'expid_19000101_m2_1_POST': ['19000101_m2_1'],
            'expid_19000101_m2_2_POST': ['19000101_m2_2'],
            'expid_19000202_m1_1_POST': ['19000202_m1_1'], 'expid_19000202_m1_2_POST': ['19000202_m1_2'], 'expid_19000202_m2_1_POST': ['19000202_m2_1'],
            'expid_19000202_m2_2_POST': ['19000202_m2_2'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1_1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1_2'], 'expid_19000101_m2_1_CLEAN': ['19000101_m2_1'],
            'expid_19000101_m2_2_CLEAN': ['19000101_m2_2'],
            'expid_19000202_m1_1_CLEAN': ['19000202_m1_1'], 'expid_19000202_m1_2_CLEAN': ['19000202_m1_2'], 'expid_19000202_m2_1_CLEAN': ['19000202_m2_1'],
            'expid_19000202_m2_2_CLEAN': ['19000202_m2_2']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.chunk is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('chunk', self.job_list.get_job_list(), self.job_list)
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_group_by_split(self):
        for date in ['19000101', '19000202']:
            for member in ['m1', 'm2']:
                for chunk in [1, 2]:
                    for split in [1, 2]:
                        job = self._createDummyJob('expid_' + date + '_' + member + '_' + str(chunk) + '_' + str(split) + '_CMORATM',
                                                   Status.WAITING, date, member, chunk, split)
                        self.job_list.get_job_list().append(job)

        groups_dict = dict()

        groups_dict['status'] = {
            'expid_19000101_m1_1_CMORATM': Status.WAITING,
            'expid_19000101_m1_2_CMORATM': Status.WAITING,
            'expid_19000101_m2_1_CMORATM': Status.WAITING,
            'expid_19000101_m2_2_CMORATM': Status.WAITING,
            'expid_19000202_m1_1_CMORATM': Status.WAITING,
            'expid_19000202_m1_2_CMORATM': Status.WAITING,
            'expid_19000202_m2_1_CMORATM': Status.WAITING,
            'expid_19000202_m2_2_CMORATM': Status.WAITING,
        }
        
        groups_dict['jobs'] =  {
            'expid_19000101_m1_1_1_CMORATM' : ['expid_19000101_m1_1_CMORATM'],
            'expid_19000101_m1_1_2_CMORATM' : ['expid_19000101_m1_1_CMORATM'],
            'expid_19000101_m1_2_1_CMORATM' : ['expid_19000101_m1_2_CMORATM'],
            'expid_19000101_m1_2_2_CMORATM' : ['expid_19000101_m1_2_CMORATM'],
            'expid_19000101_m2_1_1_CMORATM': ['expid_19000101_m2_1_CMORATM'],
            'expid_19000101_m2_1_2_CMORATM': ['expid_19000101_m2_1_CMORATM'],
            'expid_19000101_m2_2_1_CMORATM': ['expid_19000101_m2_2_CMORATM'],
            'expid_19000101_m2_2_2_CMORATM': ['expid_19000101_m2_2_CMORATM'],
            'expid_19000202_m1_1_1_CMORATM': ['expid_19000202_m1_1_CMORATM'],
            'expid_19000202_m1_1_2_CMORATM': ['expid_19000202_m1_1_CMORATM'],
            'expid_19000202_m1_2_1_CMORATM': ['expid_19000202_m1_2_CMORATM'],
            'expid_19000202_m1_2_2_CMORATM': ['expid_19000202_m1_2_CMORATM'],
            'expid_19000202_m2_1_1_CMORATM': ['expid_19000202_m2_1_CMORATM'],
            'expid_19000202_m2_1_2_CMORATM': ['expid_19000202_m2_1_CMORATM'],
            'expid_19000202_m2_2_1_CMORATM': ['expid_19000202_m2_2_CMORATM'],
            'expid_19000202_m2_2_2_CMORATM': ['expid_19000202_m2_2_CMORATM']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        job_grouping = JobGrouping('split', self.job_list.get_job_list(), self.job_list)
        self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_automatic_grouping_all(self):
        groups_dict = dict()

        groups_dict['status'] = {'19000101': Status.WAITING, '19000202': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI': ['19000101'], 'expid_19000101_m2_INI': ['19000101'], 'expid_19000202_m1_INI': ['19000202'], 'expid_19000202_m2_INI': ['19000202'],

            'expid_19000101_m1_1_SIM': ['19000101'], 'expid_19000101_m1_2_SIM': ['19000101'], 'expid_19000101_m2_1_SIM': ['19000101'],
            'expid_19000101_m2_2_SIM': ['19000101'],
            'expid_19000202_m1_1_SIM': ['19000202'], 'expid_19000202_m1_2_SIM': ['19000202'], 'expid_19000202_m2_1_SIM': ['19000202'],
            'expid_19000202_m2_2_SIM': ['19000202'],

            'expid_19000101_m1_1_POST': ['19000101'], 'expid_19000101_m1_2_POST': ['19000101'], 'expid_19000101_m2_1_POST': ['19000101'],
            'expid_19000101_m2_2_POST': ['19000101'],
            'expid_19000202_m1_1_POST': ['19000202'], 'expid_19000202_m1_2_POST': ['19000202'], 'expid_19000202_m2_1_POST': ['19000202'],
            'expid_19000202_m2_2_POST': ['19000202'],

            'expid_19000101_m1_1_CLEAN': ['19000101'], 'expid_19000101_m1_2_CLEAN': ['19000101'], 'expid_19000101_m2_1_CLEAN': ['19000101'],
            'expid_19000101_m2_2_CLEAN': ['19000101'],
            'expid_19000202_m1_1_CLEAN': ['19000202'], 'expid_19000202_m1_2_CLEAN': ['19000202'], 'expid_19000202_m2_1_CLEAN': ['19000202'],
            'expid_19000202_m2_2_CLEAN': ['19000202']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        '''side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.date is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):'''
        job_grouping = JobGrouping('automatic', self.job_list.get_job_list(), self.job_list)
        self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_automatic_grouping_not_ini(self):
        self.job_list.get_job_by_name('expid_19000101_m1_INI').status = Status.READY
        self.job_list.get_job_by_name('expid_19000101_m2_INI').status = Status.READY
        self.job_list.get_job_by_name('expid_19000202_m1_INI').status = Status.READY
        self.job_list.get_job_by_name('expid_19000202_m2_INI').status = Status.READY

        groups_dict = dict()

        groups_dict['status'] = {'19000101': Status.WAITING, '19000202': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_1_SIM': ['19000101'], 'expid_19000101_m1_2_SIM': ['19000101'], 'expid_19000101_m2_1_SIM': ['19000101'],
            'expid_19000101_m2_2_SIM': ['19000101'],
            'expid_19000202_m1_1_SIM': ['19000202'], 'expid_19000202_m1_2_SIM': ['19000202'], 'expid_19000202_m2_1_SIM': ['19000202'],
            'expid_19000202_m2_2_SIM': ['19000202'],

            'expid_19000101_m1_1_POST': ['19000101'], 'expid_19000101_m1_2_POST': ['19000101'], 'expid_19000101_m2_1_POST': ['19000101'],
            'expid_19000101_m2_2_POST': ['19000101'],
            'expid_19000202_m1_1_POST': ['19000202'], 'expid_19000202_m1_2_POST': ['19000202'], 'expid_19000202_m2_1_POST': ['19000202'],
            'expid_19000202_m2_2_POST': ['19000202'],

            'expid_19000101_m1_1_CLEAN': ['19000101'], 'expid_19000101_m1_2_CLEAN': ['19000101'], 'expid_19000101_m2_1_CLEAN': ['19000101'],
            'expid_19000101_m2_2_CLEAN': ['19000101'],
            'expid_19000202_m1_1_CLEAN': ['19000202'], 'expid_19000202_m1_2_CLEAN': ['19000202'], 'expid_19000202_m2_1_CLEAN': ['19000202'],
            'expid_19000202_m2_2_CLEAN': ['19000202']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.date is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('automatic', self.job_list.get_job_list(), self.job_list)
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_automatic_grouping_splits(self,):
        for date in ['19000101', '19000202']:
            for member in ['m1', 'm2']:
                for chunk in [1, 2]:
                    for split in [1, 2]:
                        job = self._createDummyJob(
                            'expid_' + date + '_' + member + '_' + str(chunk) + '_' + str(split) + '_CMORATM',
                            Status.WAITING, date, member, chunk, split)
                        self.job_list.get_job_list().append(job)

        groups_dict = dict()

        groups_dict['status'] = {'19000101': Status.WAITING, '19000202': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI': ['19000101'], 'expid_19000101_m2_INI': ['19000101'], 'expid_19000202_m1_INI': ['19000202'], 'expid_19000202_m2_INI': ['19000202'],

            'expid_19000101_m1_1_SIM': ['19000101'], 'expid_19000101_m1_2_SIM': ['19000101'], 'expid_19000101_m2_1_SIM': ['19000101'],
            'expid_19000101_m2_2_SIM': ['19000101'],
            'expid_19000202_m1_1_SIM': ['19000202'], 'expid_19000202_m1_2_SIM': ['19000202'], 'expid_19000202_m2_1_SIM': ['19000202'],
            'expid_19000202_m2_2_SIM': ['19000202'],

            'expid_19000101_m1_1_POST': ['19000101'], 'expid_19000101_m1_2_POST': ['19000101'], 'expid_19000101_m2_1_POST': ['19000101'],
            'expid_19000101_m2_2_POST': ['19000101'],
            'expid_19000202_m1_1_POST': ['19000202'], 'expid_19000202_m1_2_POST': ['19000202'], 'expid_19000202_m2_1_POST': ['19000202'],
            'expid_19000202_m2_2_POST': ['19000202'],

            'expid_19000101_m1_1_CLEAN': ['19000101'], 'expid_19000101_m1_2_CLEAN': ['19000101'], 'expid_19000101_m2_1_CLEAN': ['19000101'],
            'expid_19000101_m2_2_CLEAN': ['19000101'],
            'expid_19000202_m1_1_CLEAN': ['19000202'], 'expid_19000202_m1_2_CLEAN': ['19000202'], 'expid_19000202_m2_1_CLEAN': ['19000202'],
            'expid_19000202_m2_2_CLEAN': ['19000202'],

            'expid_19000101_m1_1_1_CMORATM': ['19000101'],
            'expid_19000101_m1_1_2_CMORATM': ['19000101'],
            'expid_19000101_m1_2_1_CMORATM': ['19000101'],
            'expid_19000101_m1_2_2_CMORATM': ['19000101'],
            'expid_19000101_m2_1_1_CMORATM': ['19000101'],
            'expid_19000101_m2_1_2_CMORATM': ['19000101'],
            'expid_19000101_m2_2_1_CMORATM': ['19000101'],
            'expid_19000101_m2_2_2_CMORATM': ['19000101'],
            'expid_19000202_m1_1_1_CMORATM': ['19000202'],
            'expid_19000202_m1_1_2_CMORATM': ['19000202'],
            'expid_19000202_m1_2_1_CMORATM': ['19000202'],
            'expid_19000202_m1_2_2_CMORATM': ['19000202'],
            'expid_19000202_m2_1_1_CMORATM': ['19000202'],
            'expid_19000202_m2_1_2_CMORATM': ['19000202'],
            'expid_19000202_m2_2_1_CMORATM': ['19000202'],
            'expid_19000202_m2_2_2_CMORATM': ['19000202']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.date is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('automatic', self.job_list.get_job_list(), self.job_list)
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_automatic_grouping_different_status_member(self):
        self.job_list.get_job_by_name('expid_19000101_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m2_INI').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_2_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_2_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_CLEAN').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_2_CLEAN').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m2_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_CLEAN').status = Status.RUNNING

        self.job_list.get_job_by_name('expid_19000101_m2_2_SIM').status = Status.READY

        groups_dict = dict()

        groups_dict['status'] = {'19000101_m1': Status.COMPLETED, '19000202': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI' : ['19000101_m1'],

            'expid_19000101_m1_1_SIM': ['19000101_m1'], 'expid_19000101_m1_2_SIM': ['19000101_m1'],
            'expid_19000202_m1_1_SIM': ['19000202'], 'expid_19000202_m1_2_SIM': ['19000202'], 'expid_19000202_m2_1_SIM': ['19000202'],
            'expid_19000202_m2_2_SIM': ['19000202'],

            'expid_19000101_m1_1_POST': ['19000101_m1'], 'expid_19000101_m1_2_POST': ['19000101_m1'],
            'expid_19000202_m1_1_POST': ['19000202'], 'expid_19000202_m1_2_POST': ['19000202'], 'expid_19000202_m2_1_POST': ['19000202'],
            'expid_19000202_m2_2_POST': ['19000202'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1'],
            'expid_19000202_m1_1_CLEAN': ['19000202'], 'expid_19000202_m1_2_CLEAN': ['19000202'], 'expid_19000202_m2_1_CLEAN': ['19000202'],
            'expid_19000202_m2_2_CLEAN': ['19000202']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.date is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('automatic', self.job_list.get_job_list(), self.job_list)
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_automatic_grouping_different_status_chunk(self):
        self.job_list.get_job_by_name('expid_19000101_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m2_INI').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_CLEAN').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_2_SIM').status = Status.READY

        self.job_list.get_job_by_name('expid_19000101_m2_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_CLEAN').status = Status.RUNNING

        self.job_list.get_job_by_name('expid_19000101_m2_2_SIM').status = Status.READY

        groups_dict = dict()

        groups_dict['status'] = {'19000101_m1_1': Status.COMPLETED, '19000202': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_1_SIM': ['19000101_m1_1'],
            'expid_19000202_m1_1_SIM': ['19000202'], 'expid_19000202_m1_2_SIM': ['19000202'], 'expid_19000202_m2_1_SIM': ['19000202'],
            'expid_19000202_m2_2_SIM': ['19000202'],

            'expid_19000101_m1_1_POST': ['19000101_m1_1'],
            'expid_19000202_m1_1_POST': ['19000202'], 'expid_19000202_m1_2_POST': ['19000202'], 'expid_19000202_m2_1_POST': ['19000202'],
            'expid_19000202_m2_2_POST': ['19000202'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1_1'],
            'expid_19000202_m1_1_CLEAN': ['19000202'], 'expid_19000202_m1_2_CLEAN': ['19000202'], 'expid_19000202_m2_1_CLEAN': ['19000202'],
            'expid_19000202_m2_2_CLEAN': ['19000202']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.date is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('automatic', self.job_list.get_job_list(), self.job_list)
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_group_by_member_expand_running(self):
        self.job_list.get_job_by_name('expid_19000101_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m2_INI').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_CLEAN').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_2_SIM').status = Status.READY

        self.job_list.get_job_by_name('expid_19000101_m2_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_CLEAN').status = Status.RUNNING

        self.job_list.get_job_by_name('expid_19000101_m2_2_SIM').status = Status.READY

        groups_dict = dict()

        groups_dict['status'] = {'19000101_m1': Status.READY, '19000202_m1': Status.WAITING,
                                 '19000202_m2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI': ['19000101_m1'], 'expid_19000202_m1_INI': ['19000202_m1'],
            'expid_19000202_m2_INI': ['19000202_m2'],

            'expid_19000101_m1_1_SIM': ['19000101_m1'], 'expid_19000101_m1_2_SIM': ['19000101_m1'],
            'expid_19000202_m1_1_SIM': ['19000202_m1'], 'expid_19000202_m1_2_SIM': ['19000202_m1'], 'expid_19000202_m2_1_SIM': ['19000202_m2'],
            'expid_19000202_m2_2_SIM': ['19000202_m2'],

            'expid_19000101_m1_1_POST': ['19000101_m1'], 'expid_19000101_m1_2_POST': ['19000101_m1'],
            'expid_19000202_m1_1_POST': ['19000202_m1'], 'expid_19000202_m1_2_POST': ['19000202_m1'], 'expid_19000202_m2_1_POST': ['19000202_m2'],
            'expid_19000202_m2_2_POST': ['19000202_m2'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1'],
            'expid_19000202_m1_1_CLEAN': ['19000202_m1'], 'expid_19000202_m1_2_CLEAN': ['19000202_m1'], 'expid_19000202_m2_1_CLEAN': ['19000202_m2'],
            'expid_19000202_m2_2_CLEAN': ['19000202_m2']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.member is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('member', self.job_list.get_job_list(), self.job_list, expanded_status=[Status.RUNNING])
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_group_by_chunk_expand_failed_running(self):
        self.job_list.get_job_by_name('expid_19000101_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m2_INI').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_CLEAN').status = Status.FAILED

        self.job_list.get_job_by_name('expid_19000101_m1_2_SIM').status = Status.READY

        self.job_list.get_job_by_name('expid_19000101_m2_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_CLEAN').status = Status.RUNNING

        self.job_list.get_job_by_name('expid_19000101_m2_2_SIM').status = Status.READY

        groups_dict = dict()

        groups_dict['status'] = {'19000101_m1_2': Status.READY, '19000101_m2_2': Status.READY,
                                 '19000202_m1_1': Status.WAITING, '19000202_m1_2': Status.WAITING,
                                 '19000202_m2_1': Status.WAITING, '19000202_m2_2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_2_SIM': ['19000101_m1_2'],
            'expid_19000101_m2_2_SIM': ['19000101_m2_2'],
            'expid_19000202_m1_1_SIM': ['19000202_m1_1'], 'expid_19000202_m1_2_SIM': ['19000202_m1_2'], 'expid_19000202_m2_1_SIM': ['19000202_m2_1'],
            'expid_19000202_m2_2_SIM': ['19000202_m2_2'],

            'expid_19000101_m1_2_POST': ['19000101_m1_2'],
            'expid_19000101_m2_2_POST': ['19000101_m2_2'],
            'expid_19000202_m1_1_POST': ['19000202_m1_1'], 'expid_19000202_m1_2_POST': ['19000202_m1_2'], 'expid_19000202_m2_1_POST': ['19000202_m2_1'],
            'expid_19000202_m2_2_POST': ['19000202_m2_2'],

            'expid_19000101_m1_2_CLEAN': ['19000101_m1_2'],
            'expid_19000101_m2_2_CLEAN': ['19000101_m2_2'],
            'expid_19000202_m1_1_CLEAN': ['19000202_m1_1'], 'expid_19000202_m1_2_CLEAN': ['19000202_m1_2'], 'expid_19000202_m2_1_CLEAN': ['19000202_m2_1'],
            'expid_19000202_m2_2_CLEAN': ['19000202_m2_2']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.chunk is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('chunk', self.job_list.get_job_list(), self.job_list, expanded_status=[Status.RUNNING, Status.FAILED])
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_group_by_member_expand(self):
        self.job_list.get_job_by_name('expid_19000101_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m2_INI').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_CLEAN').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_2_SIM').status = Status.READY

        self.job_list.get_job_by_name('expid_19000101_m2_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_CLEAN').status = Status.RUNNING

        self.job_list.get_job_by_name('expid_19000101_m2_2_SIM').status = Status.READY

        groups_dict = dict()

        groups_dict['status'] = {'19000101_m1': Status.READY}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI': ['19000101_m1'],

            'expid_19000101_m1_1_SIM': ['19000101_m1'], 'expid_19000101_m1_2_SIM': ['19000101_m1'],

            'expid_19000101_m1_1_POST': ['19000101_m1'], 'expid_19000101_m1_2_POST': ['19000101_m1'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1'],
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.member is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('member', self.job_list.get_job_list(), self.job_list,
                                   expand_list="[ 19000101 [m2] 19000202 [m1 m2] ]")
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_group_by_member_expand_and_running(self, *patches):
        self.job_list.get_job_by_name('expid_19000101_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m2_INI').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_CLEAN').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_2_SIM').status = Status.READY

        self.job_list.get_job_by_name('expid_19000101_m2_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_CLEAN').status = Status.RUNNING

        self.job_list.get_job_by_name('expid_19000101_m2_2_SIM').status = Status.READY

        self.job_list.get_job_by_name('expid_19000202_m1_1_SIM').status = Status.RUNNING

        groups_dict = dict()

        groups_dict['status'] = {'19000101_m1': Status.READY}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI': ['19000101_m1'],

            'expid_19000101_m1_1_SIM': ['19000101_m1'], 'expid_19000101_m1_2_SIM': ['19000101_m1'],

            'expid_19000101_m1_1_POST': ['19000101_m1'], 'expid_19000101_m1_2_POST': ['19000101_m1'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1'],
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.member is not None:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('member', self.job_list.get_job_list(), self.job_list,
                                   expand_list="[ 19000101 [m2] 19000202 [m2] ]", expanded_status=[Status.RUNNING])
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_group_by_chunk_expand(self, *patches):
        self.job_list.get_job_by_name('expid_19000101_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m1_INI').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000202_m2_INI').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_19000101_m1_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m1_1_CLEAN').status = Status.FAILED

        self.job_list.get_job_by_name('expid_19000101_m1_2_SIM').status = Status.READY

        self.job_list.get_job_by_name('expid_19000101_m2_1_SIM').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_POST').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_19000101_m2_1_CLEAN').status = Status.RUNNING

        groups_dict = dict()

        groups_dict['status'] = {'19000101_m1_1': Status.FAILED, '19000101_m1_2': Status.READY, '19000101_m2_1': Status.RUNNING, '19000202_m2_2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_1_SIM': ['19000101_m1_1'], 'expid_19000101_m1_2_SIM': ['19000101_m1_2'], 'expid_19000101_m2_1_SIM': ['19000101_m2_1'],
            'expid_19000202_m2_2_SIM': ['19000202_m2_2'],

            'expid_19000101_m1_1_POST': ['19000101_m1_1'], 'expid_19000101_m1_2_POST': ['19000101_m1_2'], 'expid_19000101_m2_1_POST': ['19000101_m2_1'],
            'expid_19000202_m2_2_POST': ['19000202_m2_2'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1_1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1_2'], 'expid_19000101_m2_1_CLEAN': ['19000101_m2_1'],
            'expid_19000202_m2_2_CLEAN': ['19000202_m2_2']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        job_grouping = JobGrouping('chunk', self.job_list.get_job_list(), self.job_list,
                               expand_list="[ 19000101 [m2 [2] ] 19000202 [m1 [1 2] m2 [1] ] ]")
        self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_synchronize_member_group_member(self):
        for date in ['19000101', '19000202']:
            for chunk in [1, 2]:
                job = self._createDummyJob('expid_' + date + '_' + str(chunk) + '_ASIM',
                                           Status.WAITING, date, None, chunk)
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

                self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'19000101_m1': Status.WAITING,
                                 '19000101_m2': Status.WAITING,
                                 '19000202_m1': Status.WAITING,
                                 '19000202_m2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI': ['19000101_m1'], 'expid_19000101_m2_INI': ['19000101_m2'], 'expid_19000202_m1_INI': ['19000202_m1'],
            'expid_19000202_m2_INI': ['19000202_m2'],
            'expid_19000101_m1_1_SIM': ['19000101_m1'], 'expid_19000101_m1_2_SIM': ['19000101_m1'], 'expid_19000101_m2_1_SIM': ['19000101_m2'],
            'expid_19000101_m2_2_SIM': ['19000101_m2'],
            'expid_19000202_m1_1_SIM': ['19000202_m1'], 'expid_19000202_m1_2_SIM': ['19000202_m1'], 'expid_19000202_m2_1_SIM': ['19000202_m2'],
            'expid_19000202_m2_2_SIM': ['19000202_m2'],

            'expid_19000101_m1_1_POST': ['19000101_m1'], 'expid_19000101_m1_2_POST': ['19000101_m1'], 'expid_19000101_m2_1_POST': ['19000101_m2'],
            'expid_19000101_m2_2_POST': ['19000101_m2'],
            'expid_19000202_m1_1_POST': ['19000202_m1'], 'expid_19000202_m1_2_POST': ['19000202_m1'], 'expid_19000202_m2_1_POST': ['19000202_m2'],
            'expid_19000202_m2_2_POST': ['19000202_m2'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1'], 'expid_19000101_m2_1_CLEAN': ['19000101_m2'],
            'expid_19000101_m2_2_CLEAN': ['19000101_m2'],
            'expid_19000202_m1_1_CLEAN': ['19000202_m1'], 'expid_19000202_m1_2_CLEAN': ['19000202_m1'], 'expid_19000202_m2_1_CLEAN': ['19000202_m2'],
            'expid_19000202_m2_2_CLEAN': ['19000202_m2'],

            'expid_19000101_1_ASIM' : ['19000101_m1', '19000101_m2'], 'expid_19000101_2_ASIM' : ['19000101_m1', '19000101_m2'],
            'expid_19000202_1_ASIM': ['19000202_m1', '19000202_m2'], 'expid_19000202_2_ASIM' : ['19000202_m1', '19000202_m2']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        job_grouping = JobGrouping('member', self.job_list.get_job_list(), self.job_list)
        self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_synchronize_member_group_chunk(self):
        for date in ['19000101', '19000202']:
            for chunk in [1, 2]:
                job = self._createDummyJob('expid_' + date + '_' + str(chunk) + '_ASIM',
                                           Status.WAITING, date, None, chunk)
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

                self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'19000101_m1_1': Status.WAITING, '19000101_m1_2': Status.WAITING,
                                 '19000101_m2_1': Status.WAITING, '19000101_m2_2': Status.WAITING,
                                 '19000202_m1_1': Status.WAITING, '19000202_m1_2': Status.WAITING,
                                 '19000202_m2_1': Status.WAITING, '19000202_m2_2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_1_SIM': ['19000101_m1_1'], 'expid_19000101_m1_2_SIM': ['19000101_m1_2'], 'expid_19000101_m2_1_SIM': ['19000101_m2_1'],
            'expid_19000101_m2_2_SIM': ['19000101_m2_2'],
            'expid_19000202_m1_1_SIM': ['19000202_m1_1'], 'expid_19000202_m1_2_SIM': ['19000202_m1_2'], 'expid_19000202_m2_1_SIM': ['19000202_m2_1'],
            'expid_19000202_m2_2_SIM': ['19000202_m2_2'],

            'expid_19000101_m1_1_POST': ['19000101_m1_1'], 'expid_19000101_m1_2_POST': ['19000101_m1_2'], 'expid_19000101_m2_1_POST': ['19000101_m2_1'],
            'expid_19000101_m2_2_POST': ['19000101_m2_2'],
            'expid_19000202_m1_1_POST': ['19000202_m1_1'], 'expid_19000202_m1_2_POST': ['19000202_m1_2'], 'expid_19000202_m2_1_POST': ['19000202_m2_1'],
            'expid_19000202_m2_2_POST': ['19000202_m2_2'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1_1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1_2'], 'expid_19000101_m2_1_CLEAN': ['19000101_m2_1'],
            'expid_19000101_m2_2_CLEAN': ['19000101_m2_2'],
            'expid_19000202_m1_1_CLEAN': ['19000202_m1_1'], 'expid_19000202_m1_2_CLEAN': ['19000202_m1_2'], 'expid_19000202_m2_1_CLEAN': ['19000202_m2_1'],
            'expid_19000202_m2_2_CLEAN': ['19000202_m2_2'],

            'expid_19000101_1_ASIM' : ['19000101_m1_1', '19000101_m2_1'], 'expid_19000101_2_ASIM' : ['19000101_m1_2', '19000101_m2_2'],
            'expid_19000202_1_ASIM': ['19000202_m1_1', '19000202_m2_1'], 'expid_19000202_2_ASIM' : ['19000202_m1_2', '19000202_m2_2']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        job_grouping = JobGrouping('chunk', self.job_list.get_job_list(), self.job_list)
        self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_synchronize_member_group_date(self):
        for date in ['19000101', '19000202']:
            for chunk in [1, 2]:
                job = self._createDummyJob('expid_' + date + '_' + str(chunk) + '_ASIM',
                                           Status.WAITING, date, None, chunk)
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

                self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'19000101': Status.WAITING,
                                 '19000202': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI': ['19000101'], 'expid_19000101_m2_INI': ['19000101'], 'expid_19000202_m1_INI': ['19000202'],
            'expid_19000202_m2_INI': ['19000202'],
            'expid_19000101_m1_1_SIM': ['19000101'], 'expid_19000101_m1_2_SIM': ['19000101'], 'expid_19000101_m2_1_SIM': ['19000101'],
            'expid_19000101_m2_2_SIM': ['19000101'],
            'expid_19000202_m1_1_SIM': ['19000202'], 'expid_19000202_m1_2_SIM': ['19000202'], 'expid_19000202_m2_1_SIM': ['19000202'],
            'expid_19000202_m2_2_SIM': ['19000202'],

            'expid_19000101_m1_1_POST': ['19000101'], 'expid_19000101_m1_2_POST': ['19000101'], 'expid_19000101_m2_1_POST': ['19000101'],
            'expid_19000101_m2_2_POST': ['19000101'],
            'expid_19000202_m1_1_POST': ['19000202'], 'expid_19000202_m1_2_POST': ['19000202'], 'expid_19000202_m2_1_POST': ['19000202'],
            'expid_19000202_m2_2_POST': ['19000202'],

            'expid_19000101_m1_1_CLEAN': ['19000101'], 'expid_19000101_m1_2_CLEAN': ['19000101'], 'expid_19000101_m2_1_CLEAN': ['19000101'],
            'expid_19000101_m2_2_CLEAN': ['19000101'],
            'expid_19000202_m1_1_CLEAN': ['19000202'], 'expid_19000202_m1_2_CLEAN': ['19000202'], 'expid_19000202_m2_1_CLEAN': ['19000202'],
            'expid_19000202_m2_2_CLEAN': ['19000202'],

            'expid_19000101_1_ASIM': ['19000101'], 'expid_19000101_2_ASIM': ['19000101'],
            'expid_19000202_1_ASIM': ['19000202'], 'expid_19000202_2_ASIM': ['19000202']
        }

        job_grouping = JobGrouping('date', self.job_list.get_job_list(), self.job_list)
        self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_synchronize_date_group_member(self):
        for chunk in [1, 2]:
            job = self._createDummyJob('expid_' + str(chunk) + '_ASIM',
                                           Status.WAITING, None, None, chunk)
            for date in ['19000101', '19000202']:
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

            self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'19000101_m1': Status.WAITING,
                                 '19000101_m2': Status.WAITING,
                                 '19000202_m1': Status.WAITING,
                                 '19000202_m2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI': ['19000101_m1'], 'expid_19000101_m2_INI': ['19000101_m2'], 'expid_19000202_m1_INI': ['19000202_m1'],
            'expid_19000202_m2_INI': ['19000202_m2'],
            'expid_19000101_m1_1_SIM': ['19000101_m1'], 'expid_19000101_m1_2_SIM': ['19000101_m1'], 'expid_19000101_m2_1_SIM': ['19000101_m2'],
            'expid_19000101_m2_2_SIM': ['19000101_m2'],
            'expid_19000202_m1_1_SIM': ['19000202_m1'], 'expid_19000202_m1_2_SIM': ['19000202_m1'], 'expid_19000202_m2_1_SIM': ['19000202_m2'],
            'expid_19000202_m2_2_SIM': ['19000202_m2'],

            'expid_19000101_m1_1_POST': ['19000101_m1'], 'expid_19000101_m1_2_POST': ['19000101_m1'], 'expid_19000101_m2_1_POST': ['19000101_m2'],
            'expid_19000101_m2_2_POST': ['19000101_m2'],
            'expid_19000202_m1_1_POST': ['19000202_m1'], 'expid_19000202_m1_2_POST': ['19000202_m1'], 'expid_19000202_m2_1_POST': ['19000202_m2'],
            'expid_19000202_m2_2_POST': ['19000202_m2'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1'], 'expid_19000101_m2_1_CLEAN': ['19000101_m2'],
            'expid_19000101_m2_2_CLEAN': ['19000101_m2'],
            'expid_19000202_m1_1_CLEAN': ['19000202_m1'], 'expid_19000202_m1_2_CLEAN': ['19000202_m1'], 'expid_19000202_m2_1_CLEAN': ['19000202_m2'],
            'expid_19000202_m2_2_CLEAN': ['19000202_m2'],

            'expid_1_ASIM' : ['19000101_m1', '19000101_m2', '19000202_m1', '19000202_m2'], 'expid_2_ASIM' : ['19000101_m1', '19000101_m2', '19000202_m1', '19000202_m2']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.date is None and job.chunk is not None:
                side_effect.append('19000101')
                side_effect.append('19000101')
                side_effect.append('19000101')
                side_effect.append('19000202')
                side_effect.append('19000202')
                side_effect.append('19000202')
            else:
                side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('member', self.job_list.get_job_list(), self.job_list)
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_synchronize_date_group_chunk(self):
        for chunk in [1, 2]:
            job = self._createDummyJob('expid_' + str(chunk) + '_ASIM',
                                           Status.WAITING, None, None, chunk)
            for date in ['19000101', '19000202']:
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

            self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'19000101_m1_1': Status.WAITING, '19000101_m1_2': Status.WAITING,
                                 '19000101_m2_1': Status.WAITING, '19000101_m2_2': Status.WAITING,
                                 '19000202_m1_1': Status.WAITING, '19000202_m1_2': Status.WAITING,
                                 '19000202_m2_1': Status.WAITING, '19000202_m2_2': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_1_SIM': ['19000101_m1_1'], 'expid_19000101_m1_2_SIM': ['19000101_m1_2'], 'expid_19000101_m2_1_SIM': ['19000101_m2_1'],
            'expid_19000101_m2_2_SIM': ['19000101_m2_2'],
            'expid_19000202_m1_1_SIM': ['19000202_m1_1'], 'expid_19000202_m1_2_SIM': ['19000202_m1_2'], 'expid_19000202_m2_1_SIM': ['19000202_m2_1'],
            'expid_19000202_m2_2_SIM': ['19000202_m2_2'],

            'expid_19000101_m1_1_POST': ['19000101_m1_1'], 'expid_19000101_m1_2_POST': ['19000101_m1_2'], 'expid_19000101_m2_1_POST': ['19000101_m2_1'],
            'expid_19000101_m2_2_POST': ['19000101_m2_2'],
            'expid_19000202_m1_1_POST': ['19000202_m1_1'], 'expid_19000202_m1_2_POST': ['19000202_m1_2'], 'expid_19000202_m2_1_POST': ['19000202_m2_1'],
            'expid_19000202_m2_2_POST': ['19000202_m2_2'],

            'expid_19000101_m1_1_CLEAN': ['19000101_m1_1'], 'expid_19000101_m1_2_CLEAN': ['19000101_m1_2'], 'expid_19000101_m2_1_CLEAN': ['19000101_m2_1'],
            'expid_19000101_m2_2_CLEAN': ['19000101_m2_2'],
            'expid_19000202_m1_1_CLEAN': ['19000202_m1_1'], 'expid_19000202_m1_2_CLEAN': ['19000202_m1_2'], 'expid_19000202_m2_1_CLEAN': ['19000202_m2_1'],
            'expid_19000202_m2_2_CLEAN': ['19000202_m2_2'],

            'expid_1_ASIM' : ['19000101_m1_1', '19000101_m2_1', '19000202_m1_1', '19000202_m2_1'], 'expid_2_ASIM' : ['19000101_m1_2', '19000101_m2_2', '19000202_m1_2', '19000202_m2_2'],
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.chunk is not None:
                if job.date is None:
                    side_effect.append('19000101')
                    side_effect.append('19000101')
                    side_effect.append('19000101')
                    side_effect.append('19000202')
                    side_effect.append('19000202')
                    side_effect.append('19000202')
                else:
                    side_effect.append(date2str(job.date, ''))

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('chunk', self.job_list.get_job_list(), self.job_list)
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def test_synchronize_date_group_date(self):
        for chunk in [1, 2]:
            job = self._createDummyJob('expid_' + str(chunk) + '_ASIM',
                                           Status.WAITING, None, None, chunk)
            for date in ['19000101', '19000202']:
                for member in ['m1', 'm2']:
                    job.add_parent(
                        self.job_list.get_job_by_name('expid_' + date + '_' + member + '_' + str(chunk) + '_SIM'))

            self.job_list.get_job_list().append(job)

        groups_dict = dict()
        groups_dict['status'] = {'19000101': Status.WAITING,
                                 '19000202': Status.WAITING}
        groups_dict['jobs'] = {
            'expid_19000101_m1_INI': ['19000101'], 'expid_19000101_m2_INI': ['19000101'], 'expid_19000202_m1_INI': ['19000202'],
            'expid_19000202_m2_INI': ['19000202'],
            'expid_19000101_m1_1_SIM': ['19000101'], 'expid_19000101_m1_2_SIM': ['19000101'], 'expid_19000101_m2_1_SIM': ['19000101'],
            'expid_19000101_m2_2_SIM': ['19000101'],
            'expid_19000202_m1_1_SIM': ['19000202'], 'expid_19000202_m1_2_SIM': ['19000202'], 'expid_19000202_m2_1_SIM': ['19000202'],
            'expid_19000202_m2_2_SIM': ['19000202'],

            'expid_19000101_m1_1_POST': ['19000101'], 'expid_19000101_m1_2_POST': ['19000101'], 'expid_19000101_m2_1_POST': ['19000101'],
            'expid_19000101_m2_2_POST': ['19000101'],
            'expid_19000202_m1_1_POST': ['19000202'], 'expid_19000202_m1_2_POST': ['19000202'], 'expid_19000202_m2_1_POST': ['19000202'],
            'expid_19000202_m2_2_POST': ['19000202'],

            'expid_19000101_m1_1_CLEAN': ['19000101'], 'expid_19000101_m1_2_CLEAN': ['19000101'], 'expid_19000101_m2_1_CLEAN': ['19000101'],
            'expid_19000101_m2_2_CLEAN': ['19000101'],
            'expid_19000202_m1_1_CLEAN': ['19000202'], 'expid_19000202_m1_2_CLEAN': ['19000202'], 'expid_19000202_m2_1_CLEAN': ['19000202'],
            'expid_19000202_m2_2_CLEAN': ['19000202'],

            'expid_1_ASIM': ['19000101', '19000202'], 'expid_2_ASIM': ['19000101', '19000202']
        }

        self.job_list.get_date_list = Mock(return_value=['19000101', '19000202'])
        self.job_list.get_member_list = Mock(return_value=['m1', 'm2'])
        self.job_list.get_chunk_list = Mock(return_value=[1, 2])
        self.job_list.get_date_format = Mock(return_value='')

        side_effect = []
        for job in reversed(self.job_list.get_job_list()):
            if job.date is not None:
                side_effect.append(date2str(job.date))
            elif job.chunk is not None:
                side_effect.append('19000101')
                side_effect.append('19000202')

        with patch('autosubmit.job.job_grouping.date2str', side_effect=side_effect):
            job_grouping = JobGrouping('date', self.job_list.get_job_list(), self.job_list)
            self.assertDictEqual(job_grouping.group_jobs(), groups_dict)

    def _createDummyJob(self, name, status, date=None, member=None, chunk=None, split=None):
        job_id = randrange(1, 999)
        job = Job(name, job_id, status, 0)
        job.type = randrange(0, 2)

        job.date = parse_date(date)
        job.member = member
        job.chunk = chunk
        job.split = split

        return job

class FakeBasicConfig:
    def __init__(self):
        pass

    DB_DIR = '/dummy/db/dir'
    DB_FILE = '/dummy/db/file'
    DB_PATH = '/dummy/db/path'
    LOCAL_ROOT_DIR = '/dummy/local/root/dir'
    LOCAL_TMP_DIR = '/dummy/local/temp/dir'
    LOCAL_PROJ_DIR = '/dummy/local/proj/dir'
    DEFAULT_PLATFORMS_CONF = ''
    DEFAULT_JOBS_CONF = ''






#!/usr/bin/env python

# Copyright 2015 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

"""
Module containing functions to manage autosubmit's experiments.
"""
import string
import autosubmit.database.db_common as db_common
from bscearth.utils.log import Log


def new_experiment(description, version, test=False, operational=False):
    """
    Stores a new experiment on the database and generates its identifier

    :param version: autosubmit version associated to the experiment
    :type version: str
    :param test: flag for test experiments
    :type test: bool
    :param operational: flag for operational experiments
    :type operational: bool
    :param description: experiment's description
    :type description: str
    :return: experiment id for the new experiment
    :rtype: str
    """
    if test:
        last_exp_name = db_common.last_name_used(True)
    elif operational:
        last_exp_name = db_common.last_name_used(False, True)
    else:
        last_exp_name = db_common.last_name_used()
    if last_exp_name == '':
        return ''
    if last_exp_name == 'empty':
        if test:
            # test identifier restricted also to 4 characters.
            new_name = 't000'
        elif operational:
            # operational identifier restricted also to 4 characters.
            new_name = 'o000'
        else:
            new_name = 'a000'
    else:
        new_name = next_experiment_id(last_exp_name)
        if new_name == '':
            return ''
    while db_common.check_experiment_exists(new_name, False):
        new_name = next_experiment_id(new_name)
        if new_name == '':
            return ''
    if not db_common.save_experiment(new_name, description, version):
        return ''
    Log.info('The new experiment "{0}" has been registered.', new_name)
    return new_name


def copy_experiment(experiment_id, description, version, test=False, operational=False):
    """
    Creates a new experiment by copying an existing experiment

    :param version: experiment's associated autosubmit version
    :type version: str
    :param experiment_id: identifier of experiment to copy
    :type experiment_id: str
    :param description: experiment's description
    :type description: str
    :param test: specifies if it is a test experiment
    :type test: bool
    :param operational: specifies if it is a operational experiment
    :type operational: bool
    :return: experiment id for the new experiment
    :rtype: str
    """
    if not db_common.check_experiment_exists(experiment_id):
        return ''
    new_name = new_experiment(description, version, test, operational)
    return new_name


def next_experiment_id(current_id):
    """
    Get next experiment identifier

    :param current_id: previous experiment identifier
    :type current_id: str
    :return: new experiment identifier
    :rtype: str
    """
    if not is_valid_experiment_id(current_id):
        return ''
    # Convert the name to base 36 in number add 1 and then encode it
    next_id = base36encode(base36decode(current_id) + 1)
    return next_id if is_valid_experiment_id(next_id) else ''


def is_valid_experiment_id(name):
    """
    Checks if it is a valid experiment identifier

    :param name: experiment identifier to check
    :type name: str
    :return: name if is valid, terminates program otherwise
    :rtype: str
    """
    name = name.lower()
    if len(name) < 4 or not name.isalnum():
        Log.error("So sorry, but the name must have at least 4 alphanumeric chars!!!")
        return False
    return True


def base36encode(number, alphabet=string.digits + string.ascii_lowercase):
    """
    Convert positive integer to a base36 string.

    :param number: number to convert
    :type number: int
    :param alphabet: set of characters to use
    :type alphabet: str
    :return: number's base36 string value
    :rtype: str
    """
    if not isinstance(number, (int, long)):
        raise TypeError('number must be an integer')

    # Special case for zero
    if number == 0:
        return '0'

    base36 = ''

    sign = ''
    if number < 0:
        sign = '-'
        number = - number

    while number > 0:
        number, i = divmod(number, len(alphabet))
        # noinspection PyAugmentAssignment
        base36 = alphabet[i] + base36

    return sign + base36.rjust(4, '0')


def base36decode(number):
    """
    Converts a base36 string to a positive integer

    :param number: base36 string to convert
    :type number: str
    :return: number's integer value
    :rtype: int
    """
    return int(number, 36)



#!/usr/bin/env python

# Copyright 2017 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.patches as mpatches
from autosubmit.experiment.statistics import ExperimentStats
from autosubmit.job.job_common import Status
from bscearth.utils.log import Log
from autosubmit.job.job import Job

# Autosubmit stats constants
RATIO = 4
MAX_JOBS_PER_PLOT = 12.0


def create_bar_diagram(experiment_id, jobs_list, general_stats, output_file, period_ini=None, period_fi=None):
    # Error prevention
    plt.close('all')
    # Stats variables definition
    num_plots = int(np.ceil(len(jobs_list) / MAX_JOBS_PER_PLOT))
    ind = np.arange(int(MAX_JOBS_PER_PLOT))
    width = 0.16
    # Creating stats figure
    fig = plt.figure(figsize=(RATIO * 4, 3 * RATIO * num_plots))
    fig.suptitle('STATS - ' + experiment_id, fontsize=24, fontweight='bold')
    # Variables initialization
    ax, ax2 = [], []
    rects = [None] * 6
    exp_stats = ExperimentStats(jobs_list, period_ini, period_fi)
    grid_spec = gridspec.GridSpec(RATIO * num_plots + 2, 1)
    for plot in range(1, num_plots + 1):
        # Calculating jobs inside the given plot
        l1 = int((plot - 1) * MAX_JOBS_PER_PLOT)
        l2 = int(plot * MAX_JOBS_PER_PLOT)
        # Building plot axis
        ax.append(fig.add_subplot(grid_spec[RATIO * plot - RATIO + 2:RATIO * plot + 1]))
        ax[plot - 1].set_ylabel('hours')
        ax[plot - 1].set_xticks(ind + width)
        ax[plot - 1].set_xticklabels([job.name for job in jobs_list[l1:l2]], rotation='vertical')
        ax[plot - 1].set_title(experiment_id, fontsize=20)
        ax[plot - 1].set_ylim(0, float(1.10 * exp_stats.max_time))
        # Axis 2
        ax2.append(ax[plot - 1].twinx())
        ax2[plot - 1].set_ylabel('# failed jobs')
        ax2[plot - 1].set_yticks(range(0, exp_stats.max_fail + 2))
        ax2[plot - 1].set_ylim(0, exp_stats.max_fail + 1)
        # Building rects
        rects[0] = ax[plot - 1].bar(ind, exp_stats.queued[l1:l2], width, color='orchid')
        rects[1] = ax[plot - 1].bar(ind + width, exp_stats.run[l1:l2], width, color='limegreen')
        rects[2] = ax2[plot - 1].bar(ind + width * 2, exp_stats.failed_jobs[l1:l2], width, color='red')
        rects[3] = ax[plot - 1].bar(ind + width * 3, exp_stats.fail_queued[l1:l2], width, color='purple')
        rects[4] = ax[plot - 1].bar(ind + width * 4, exp_stats.fail_run[l1:l2], width, color='tomato')
        rects[5] = ax[plot - 1].plot([0., width * 6 * MAX_JOBS_PER_PLOT], [exp_stats.threshold, exp_stats.threshold],
                                     "k--", label='wallclock sim')

    # Building legends subplot
    legends_plot = fig.add_subplot(grid_spec[0, 0])
    legends_plot.set_frame_on(False)
    legends_plot.axes.get_xaxis().set_visible(False)
    legends_plot.axes.get_yaxis().set_visible(False)

    # Building legends
    build_legends(legends_plot, rects, exp_stats, general_stats)

    # Saving output figure
    grid_spec.tight_layout(fig, rect=[0, 0.03, 1, 0.97])
    plt.savefig(output_file)

    create_csv_stats(exp_stats, jobs_list, output_file)

def create_csv_stats(exp_stats, jobs_list, output_file):
    job_names = [job.name for job in jobs_list]
    start_times = exp_stats.start_times
    end_times = exp_stats.end_times
    queuing_times = exp_stats.queued
    running_times = exp_stats.run

    output_file = output_file.replace('pdf', 'csv')
    with open(output_file, 'wb') as file:
        file.write("Job,Started,Ended,Queuing time (hours),Running time (hours)\n")
        for i in range(len(jobs_list)):
            file.write("{0},{1},{2},{3},{4}\n".format(job_names[i], start_times[i], end_times[i], queuing_times[i], running_times[i]))

def build_legends(plot, rects, experiment_stats, general_stats):
    # Main legend with colourful rectangles
    legend_rects = [[rect[0] for rect in rects]]
    legend_titles = [
        ['Queued (h)', 'Run (h)', 'Failed jobs (#)', 'Fail Queued (h)', 'Fail Run (h)', 'Max wallclock (h)']
    ]
    legend_locs = ["upper right"]
    legend_handlelengths = [None]

    # General stats legends, if exists
    if len(general_stats) > 0:
        legend_rects.append(get_whites_array(len(general_stats)))
        legend_titles.append([str(key) + ': ' + str(value) for key, value in general_stats])
        legend_locs.append("upper center")
        legend_handlelengths.append(0)

    # Total stats legend
    legend_rects.append(get_whites_array(len(experiment_stats.totals)))
    legend_titles.append(experiment_stats.totals)
    legend_locs.append("upper left")
    legend_handlelengths.append(0)

    # Creating the legends
    legends = create_legends(plot, legend_rects, legend_titles, legend_locs, legend_handlelengths)
    for legend in legends:
        plt.gca().add_artist(legend)


def create_legends(plot, rects, titles, locs, handlelengths):
    legends = []
    for i in xrange(len(rects)):
        legends.append(create_legend(plot, rects[i], titles[i], locs[i], handlelengths[i]))
    return legends


def create_legend(plot, rects, titles, loc, handlelength=None):
    return plot.legend(rects, titles, loc=loc, handlelength=handlelength)


def get_whites_array(length):
    white = mpatches.Rectangle((0, 0), 0, 0, alpha=0.0)
    return [white for _ in xrange(length)]



#!/usr/bin/env python

# Copyright 2015 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

"""
Module containing functions to manage autosubmit's database.
"""
import os
import sqlite3

from bscearth.utils.log import Log
from autosubmit.config.basicConfig import BasicConfig

CURRENT_DATABASE_VERSION = 1


def create_db(qry):
    """
    Creates a new database for autosubmit

    :param qry: query to create the new database
    :type qry: str    """

    try:
        (conn, cursor) = open_conn(False)
    except DbException as e:
        Log.error('Connection to database could not be established: {0}', e.message)
        return False

    try:
        cursor.executescript(qry)
    except sqlite3.Error:
        close_conn(conn, cursor)
        Log.error('The database can not be created.' + 'DB file:' + BasicConfig.DB_PATH)
        return False

    conn.commit()
    close_conn(conn, cursor)
    return True


def check_db():
    """
    Checks if database file exist

    :return: None if exists, terminates program if not
    """

    if not os.path.exists(BasicConfig.DB_PATH):
        Log.error('Some problem has happened...check the database file.' + 'DB file:' + BasicConfig.DB_PATH)
        return False
    return True


def open_conn(check_version=True):
    """
    Opens a connection to database

    :param check_version: If true, check if the database is compatible with this autosubmit version
    :type check_version: bool
    :return: connection object, cursor object
    :rtype: sqlite3.Connection, sqlite3.Cursor
    """
    conn = sqlite3.connect(BasicConfig.DB_PATH)
    cursor = conn.cursor()

    # Getting database version
    if check_version:
        try:
            cursor.execute('SELECT version '
                           'FROM db_version;')
            row = cursor.fetchone()
            version = row[0]
        except sqlite3.OperationalError:
            # If this exception is thrown it's because db_version does not exist.
            # Database is from 2.x or 3.0 beta releases
            try:
                cursor.execute('SELECT type '
                               'FROM experiment;')
                # If type field exists, it's from 2.x
                version = -1
            except sqlite3.Error:
                # If raises and error , it's from 3.0 beta releases
                version = 0

        # If database version is not the expected, update database....
        if version < CURRENT_DATABASE_VERSION:
            if not _update_database(version, cursor):
                raise DbException('Database version could not be updated')

        # ... or ask for autosubmit upgrade
        elif version > CURRENT_DATABASE_VERSION:
            Log.critical('Database version is not compatible with this autosubmit version. Please execute pip install '
                         'autosubmit --upgrade')
            raise DbException('Database version not compatible')

    return conn, cursor


def close_conn(conn, cursor):
    """
    Commits changes and close connection to database

    :param conn: connection to close
    :type conn: sqlite3.Connection
    :param cursor: cursor to close
    :type cursor: sqlite3.Cursor
    """
    conn.commit()
    cursor.close()
    conn.close()
    return


def save_experiment(name, description, version):
    """
    Stores experiment in database

    :param version:
    :type version: str
    :param name: experiment's name
    :type name: str
    :param description: experiment's description
    :type description: str
    """
    if not check_db():
        return False
    try:
        (conn, cursor) = open_conn()
    except DbException as e:
        Log.error('Connection to database could not be established: {0}', e.message)
        return False
    try:
        cursor.execute('INSERT INTO experiment (name, description, autosubmit_version) VALUES (:name, :description, '
                       ':version)',
                       {'name': name, 'description': description, 'version': version})
    except sqlite3.IntegrityError as e:
        close_conn(conn, cursor)
        Log.error('Could not register experiment: {0}'.format(e))
        return False

    conn.commit()
    close_conn(conn, cursor)
    return True


def check_experiment_exists(name, error_on_inexistence=True):
    """
    Checks if exist an experiment with the given name.

    :param error_on_inexistence: if True, adds an error log if experiment does not exists
    :type error_on_inexistence: bool
    :param name: Experiment name
    :type name: str
    :return: If experiment exists returns true, if not returns false
    :rtype: bool
    """
    if not check_db():
        return False
    try:
        (conn, cursor) = open_conn()
    except DbException as e:
        Log.error('Connection to database could not be established: {0}', e.message)
        return False
    conn.isolation_level = None

    # SQLite always return a unicode object, but we can change this
    # behaviour with the next sentence
    conn.text_factory = str
    cursor.execute('select name from experiment where name=:name', {'name': name})
    row = cursor.fetchone()
    close_conn(conn, cursor)
    if row is None:
        if error_on_inexistence:
            Log.error('The experiment name "{0}" does not exist yet!!!', name)
        return False
    return True


def get_autosubmit_version(expid):
    """
    Get the minimun autosubmit version needed for the experiment

    :param expid: Experiment name
    :type expid: str
    :return: If experiment exists returns the autosubmit version for it, if not returns None
    :rtype: str
    """
    if not check_db():
        return False

    try:
        (conn, cursor) = open_conn()
    except DbException as e:
        Log.error('Connection to database could not be established: {0}', e.message)
        return False
    conn.isolation_level = None

    # SQLite always return a unicode object, but we can change this
    # behaviour with the next sentence
    conn.text_factory = str
    cursor.execute('SELECT autosubmit_version FROM experiment WHERE name=:expid', {'expid': expid})
    row = cursor.fetchone()
    close_conn(conn, cursor)
    if row is None:
        Log.error('The experiment "{0}" does not exist yet!!!', expid)
        return None
    return row[0]


def last_name_used(test=False, operational=False):
    """
    Gets last experiment identifier used

    :param test: flag for test experiments
    :type test: bool
    :param operational: flag for operational experiments
    :type test: bool
    :return: last experiment identifier used, 'empty' if there is none
    :rtype: str
    """
    if not check_db():
        return ''
    try:
        (conn, cursor) = open_conn()
    except DbException as e:
        Log.error('Connection to database could not be established: {0}', e.message)
        return ''
    conn.text_factory = str
    if test:
        cursor.execute('SELECT name '
                       'FROM experiment '
                       'WHERE rowid=(SELECT max(rowid) FROM experiment WHERE name LIKE "t%" AND '
                       'autosubmit_version IS NOT NULL AND '
                       'NOT (autosubmit_version LIKE "%3.0.0b%"))')
    elif operational:
        cursor.execute('SELECT name '
                       'FROM experiment '
                       'WHERE rowid=(SELECT max(rowid) FROM experiment WHERE name LIKE "o%" AND '
                       'autosubmit_version IS NOT NULL AND '
                       'NOT (autosubmit_version LIKE "%3.0.0b%"))')
    else:
        cursor.execute('SELECT name '
                       'FROM experiment '
                       'WHERE rowid=(SELECT max(rowid) FROM experiment WHERE name NOT LIKE "t%" AND '
                       'name NOT LIKE "o%" AND autosubmit_version IS NOT NULL AND '
                       'NOT (autosubmit_version LIKE "%3.0.0b%"))')
    row = cursor.fetchone()
    close_conn(conn, cursor)
    if row is None:
        return 'empty'

    # If starts by number (during 3.0 beta some jobs starting with numbers where created), returns empty.
    try:
        int(row[0][0])
        return 'empty'
    except ValueError:
        return row[0]


def delete_experiment(experiment_id):
    """
    Removes experiment from database

    :param experiment_id: experiment identifier
    :type experiment_id: str
    :return: True if delete is succesful
    :rtype: bool
    """
    if not check_db():
        return False
    if not check_experiment_exists(experiment_id, False):
        return True
    try:
        (conn, cursor) = open_conn()
    except DbException as e:
        Log.error('Connection to database could not be established: {0}', e.message)
        return False
    cursor.execute('DELETE FROM experiment '
                   'WHERE name=:name', {'name': experiment_id})
    row = cursor.fetchone()
    if row is None:
        Log.debug('The experiment {0} has been deleted!!!', experiment_id)
    close_conn(conn, cursor)
    return True


def _update_database(version, cursor):
    Log.info("Autosubmit's database version is {0}. Current version is {1}. Updating...",
             version, CURRENT_DATABASE_VERSION)
    try:
        # For databases from Autosubmit 2
        if version <= -1:
            cursor.executescript('CREATE TABLE experiment_backup(id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, '
                                 'name VARCHAR NOT NULL, type VARCHAR, autosubmit_version VARCHAR, '
                                 'description VARCHAR NOT NULL, model_branch VARCHAR, template_name VARCHAR, '
                                 'template_branch VARCHAR, ocean_diagnostics_branch VARCHAR);'
                                 'INSERT INTO experiment_backup (name,type,description,model_branch,template_name,'
                                 'template_branch,ocean_diagnostics_branch) SELECT name,type,description,model_branch,'
                                 'template_name,template_branch,ocean_diagnostics_branch FROM experiment;'
                                 'UPDATE experiment_backup SET autosubmit_version = "2";'
                                 'DROP TABLE experiment;'
                                 'ALTER TABLE experiment_backup RENAME TO experiment;')
        if version <= 0:
            # Autosubmit beta version. Create db_version table
            cursor.executescript('CREATE TABLE db_version(version INTEGER NOT NULL);'
                                 'INSERT INTO db_version (version) VALUES (1);'
                                 'ALTER TABLE experiment ADD COLUMN autosubmit_version VARCHAR;'
                                 'UPDATE experiment SET autosubmit_version = "3.0.0b" '
                                 'WHERE autosubmit_version NOT NULL;')
        cursor.execute('UPDATE db_version SET version={0};'.format(CURRENT_DATABASE_VERSION))
    except sqlite3.Error as e:
        Log.critical('Can not update database: {0}', e)
        return False
    Log.info("Update completed")
    return True


class DbException(Exception):
    """
    Exception class for database errors
    """

    def __init__(self, message):
        self.message = message



#!/usr/bin/env python

# Copyright 2017 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.
try:
    # noinspection PyCompatibility
    from configparser import SafeConfigParser
except ImportError:
    # noinspection PyCompatibility
    from ConfigParser import SafeConfigParser
import json

import os
import pickle
from time import localtime, strftime
from sys import setrecursionlimit
from shutil import move

from autosubmit.job.job import Job
from bscearth.utils.log import Log
from autosubmit.job.job_dict import DicJobs
from autosubmit.job.job_utils import Dependency
from autosubmit.job.job_common import Status, Type
from bscearth.utils.date import date2str, parse_date, sum_str_hours
from autosubmit.job.job_packages import JobPackageSimple, JobPackageArray, JobPackageThread

from networkx import DiGraph
from autosubmit.job.job_utils import transitive_reduction


class JobList:
    """
    Class to manage the list of jobs to be run by autosubmit

    """

    def __init__(self, expid, config, parser_factory, job_list_persistence):
        self._persistence_path = os.path.join(config.LOCAL_ROOT_DIR, expid, "pkl")
        self._update_file = "updated_list_" + expid + ".txt"
        self._failed_file = "failed_job_list_" + expid + ".pkl"
        self._persistence_file = "job_list_" + expid
        self._job_list = list()
        self._expid = expid
        self._config = config
        self._parser_factory = parser_factory
        self._stat_val = Status()
        self._parameters = []
        self._date_list = []
        self._member_list = []
        self._chunk_list = []
        self._dic_jobs = dict()
        self._persistence = job_list_persistence
        self._graph = DiGraph()

        self.packages_dict = dict()
        self._ordered_jobs_by_date_member = dict()

        self.packages_id = dict()
        self.job_package_map = dict()
        self.sections_checked = set()

    @property
    def expid(self):
        """
        Returns the experiment identifier

        :return: experiment's identifier
        :rtype: str
        """
        return self._expid

    @property
    def graph(self):
        """
        Returns the graph

        :return: graph
        :rtype: networkx graph
        """
        return self._graph

    @graph.setter
    def graph(self, value):
        self._graph = value

    def generate(self, date_list, member_list, num_chunks, chunk_ini, parameters, date_format, default_retrials,
                 default_job_type, wrapper_type=None, wrapper_jobs=None,new=True, notransitive=False):
        """
        Creates all jobs needed for the current workflow

        :param default_job_type: default type for jobs
        :type default_job_type: str
        :param date_list: start dates
        :type date_list: list
        :param member_list: members
        :type member_list: list
        :param num_chunks: number of chunks to run
        :type num_chunks: int
        :param chunk_ini: the experiment will start by the given chunk
        :type chunk_ini: int
        :param parameters: parameters for the jobs
        :type parameters: dict
        :param date_format: option to format dates
        :type date_format: str
        :param default_retrials: default retrials for ech job
        :type default_retrials: int
        :param new: is it a new generation?
        :type new: bool
        """
        self._parameters = parameters
        self._date_list = date_list
        self._member_list = member_list

        chunk_list = range(chunk_ini, num_chunks + 1)
        self._chunk_list = chunk_list

        jobs_parser = self._get_jobs_parser()

        dic_jobs = DicJobs(self, jobs_parser, date_list, member_list, chunk_list, date_format, default_retrials)
        self._dic_jobs = dic_jobs
        priority = 0

        Log.info("Creating jobs...")
        jobs_data = dict()
        if not new:
            jobs_data = {str(row[0]): row for row in self.load()}
        self._create_jobs(dic_jobs, jobs_parser, priority, default_job_type, jobs_data)
        Log.info("Adding dependencies...")
        self._add_dependencies(date_list, member_list, chunk_list, dic_jobs, jobs_parser, self.graph)

        Log.info("Removing redundant dependencies...")
        self.update_genealogy(new, notransitive)
        for job in self._job_list:
            job.parameters = parameters

        if wrapper_type == 'vertical-mixed':
            self._ordered_jobs_by_date_member = self._create_sorted_dict_jobs(wrapper_jobs)


    @staticmethod
    def _add_dependencies(date_list, member_list, chunk_list, dic_jobs, jobs_parser, graph, option="DEPENDENCIES"):
        for job_section in jobs_parser.sections():
            Log.debug("Adding dependencies for {0} jobs".format(job_section))

            # If does not have dependencies, do nothing
            if not jobs_parser.has_option(job_section, option):
                continue

            dependencies_keys = jobs_parser.get(job_section, option).split()
            dependencies = JobList._manage_dependencies(dependencies_keys, dic_jobs)

            for job in dic_jobs.get_jobs(job_section):
                num_jobs = 1
                if isinstance(job, list):
                    num_jobs = len(job)
                for i in range(num_jobs):
                    _job = job[i] if num_jobs > 1 else job
                    JobList._manage_job_dependencies(dic_jobs, _job, date_list, member_list, chunk_list, dependencies_keys,
                                                     dependencies, graph)

    @staticmethod
    def _manage_dependencies(dependencies_keys, dic_jobs):
        dependencies = dict()
        for key in dependencies_keys:
            distance = None
            splits = None
            sign = None

            if '-' not in key and '+' not in key:
                section = key
            else:
                sign = '-' if '-' in key else '+'
                key_split = key.split(sign)
                section = key_split[0]
                distance = int(key_split[1])

            if '[' in section:
                section_name = section[0:section.find("[")]
                splits_section = int(dic_jobs.get_option(section_name, 'SPLITS', 0))
                splits = JobList._calculate_splits_dependencies(section, splits_section)
                section = section_name

            dependency_running_type = dic_jobs.get_option(section, 'RUNNING', 'once').lower()
            delay = int(dic_jobs.get_option(section, 'DELAY', -1))
            dependency = Dependency(section, distance, dependency_running_type, sign, delay, splits)
            dependencies[key] = dependency
        return dependencies

    @staticmethod
    def _calculate_splits_dependencies(section, max_splits):
        splits_list = section[section.find("[") + 1:section.find("]")]
        splits = []
        for str_split in splits_list.split(","):
            if str_split.find(":") != -1:
                numbers = str_split.split(":")
                # change this to be checked in job_common.py
                max_splits = min(int(numbers[1]), max_splits)
                for count in range(int(numbers[0]), max_splits+1):
                    splits.append(int(str(count).zfill(len(numbers[0]))))
            else:
                if int(str_split) <= max_splits:
                    splits.append(int(str_split))
        return splits

    @staticmethod
    def _manage_job_dependencies(dic_jobs, job, date_list, member_list, chunk_list, dependencies_keys, dependencies,
                                 graph):
        for key in dependencies_keys:
            dependency = dependencies[key]

            skip, (chunk, member, date) = JobList._calculate_dependency_metadata(job.chunk, chunk_list,
                                                                                 job.member, member_list,
                                                                                 job.date, date_list,
                                                                                 dependency)
            if skip:
                continue

            for parent in dic_jobs.get_jobs(dependency.section, date, member, chunk):
                if dependency.delay == -1 or chunk > dependency.delay:
                    if isinstance(parent, list):
                        if job.split is not None:
                            parent = filter(lambda _parent: _parent.split == job.split, parent)[0]
                        else:
                            if dependency.splits is not None:
                                parent = filter(lambda _parent: _parent.split in dependency.splits, parent)

                    job.add_parent(parent)
                    JobList._add_edge(graph, job, parent)

            JobList.handle_frequency_interval_dependencies(chunk, chunk_list, date, date_list, dic_jobs, job, member,
                                                           member_list, dependency.section, graph)

    @staticmethod
    def _calculate_dependency_metadata(chunk, chunk_list, member, member_list, date, date_list, dependency):
        skip = False
        if dependency.sign is '-':
            if chunk is not None and dependency.running == 'chunk':
                chunk_index = chunk_list.index(chunk)
                if chunk_index >= dependency.distance:
                    chunk = chunk_list[chunk_index - dependency.distance]
                else:
                    skip = True
            elif member is not None and dependency.running in ['chunk', 'member']:
                member_index = member_list.index(member)
                if member_index >= dependency.distance:
                    member = member_list[member_index - dependency.distance]
                else:
                    skip = True
            elif date is not None and dependency.running in ['chunk', 'member', 'startdate']:
                date_index = date_list.index(date)
                if date_index >= dependency.distance:
                    date = date_list[date_index - dependency.distance]
                else:
                    skip = True

        if dependency.sign is '+':
            if chunk is not None and dependency.running == 'chunk':
                chunk_index = chunk_list.index(chunk)
                if (chunk_index + dependency.distance) < len(chunk_list):
                    chunk = chunk_list[chunk_index + dependency.distance]
                else:  # calculating the next one possible
                    temp_distance = dependency.distance
                    while temp_distance > 0:
                        temp_distance -= 1
                        if (chunk_index + temp_distance) < len(chunk_list):
                            chunk = chunk_list[chunk_index + temp_distance]
                            break

            elif member is not None and dependency.running in ['chunk', 'member']:
                member_index = member_list.index(member)
                if (member_index + dependency.distance) < len(member_list):
                    member = member_list[member_index + dependency.distance]
                else:
                    skip = True
            elif date is not None and dependency.running in ['chunk', 'member', 'startdate']:
                date_index = date_list.index(date)
                if (date_index + dependency.distance) < len(date_list):
                    date = date_list[date_index - dependency.distance]
                else:
                    skip = True
        return skip, (chunk, member, date)

    @staticmethod
    def handle_frequency_interval_dependencies(chunk, chunk_list, date, date_list, dic_jobs, job, member, member_list,
                                               section_name, graph):
        if job.wait and job.frequency > 1:
            if job.chunk is not None:
                max_distance = (chunk_list.index(chunk) + 1) % job.frequency
                if max_distance == 0:
                    max_distance = job.frequency
                for distance in range(1, max_distance):
                    for parent in dic_jobs.get_jobs(section_name, date, member, chunk - distance):
                        job.add_parent(parent)
                        JobList._add_edge(graph, job, parent)
            elif job.member is not None:
                member_index = member_list.index(job.member)
                max_distance = (member_index + 1) % job.frequency
                if max_distance == 0:
                    max_distance = job.frequency
                for distance in range(1, max_distance, 1):
                    for parent in dic_jobs.get_jobs(section_name, date,
                                                    member_list[member_index - distance], chunk):
                        job.add_parent(parent)
                        JobList._add_edge(graph, job, parent)
            elif job.date is not None:
                date_index = date_list.index(job.date)
                max_distance = (date_index + 1) % job.frequency
                if max_distance == 0:
                    max_distance = job.frequency
                for distance in range(1, max_distance, 1):
                    for parent in dic_jobs.get_jobs(section_name, date_list[date_index - distance],
                                                    member, chunk):
                        job.add_parent(parent)
                        JobList._add_edge(graph, job, parent)

    @staticmethod
    def _add_edge(graph, job, parents):
        num_parents = 1
        if isinstance(parents, list):
            num_parents = len(parents)
        for i in range(num_parents):
            parent = parents[i] if isinstance(parents, list) else parents
            graph.add_edge(parent.name, job.name)

    @staticmethod
    def _create_jobs(dic_jobs, parser, priority, default_job_type, jobs_data=dict()):
        for section in parser.sections():
            Log.debug("Creating {0} jobs".format(section))
            dic_jobs.read_section(section, priority, default_job_type, jobs_data)
            priority += 1

    def _create_sorted_dict_jobs(self, wrapper_jobs):
        dict_jobs = dict()
        for date in self._date_list:
            dict_jobs[date] = dict()
            for member in self._member_list:
                dict_jobs[date][member] = list()
        num_chunks = len(self._chunk_list)

        filtered_jobs_list = filter(lambda job: job.section in wrapper_jobs, self._job_list)

        filtered_jobs_fake_date_member, fake_original_job_map = self._create_fake_dates_members(filtered_jobs_list)

        sections_running_type_map = dict()
        for section in wrapper_jobs.split(" "):
            sections_running_type_map[section] = self._dic_jobs.get_option(section, "RUNNING", 'once')

        for date in self._date_list:
            str_date = self._get_date(date)
            for member in self._member_list:
                sorted_jobs_list = filter(lambda job: job.name.split("_")[1] == str_date and
                                                      job.name.split("_")[2] == member, filtered_jobs_fake_date_member)

                previous_job = sorted_jobs_list[0]
                section_running_type = sections_running_type_map[previous_job.section]

                jobs_to_sort = [previous_job]
                previous_section_running_type = None

                for index in range(1, len(sorted_jobs_list) + 1):
                    if index < len(sorted_jobs_list):
                        job = sorted_jobs_list[index]

                        if previous_job.section != job.section:
                            previous_section_running_type = section_running_type
                            section_running_type = sections_running_type_map[job.section]

                    if (previous_section_running_type != None and previous_section_running_type != section_running_type) \
                      or index == len(sorted_jobs_list):

                        jobs_to_sort = sorted(jobs_to_sort, key=lambda k: (k.name.split('_')[1], (k.name.split('_')[2]),
                                                                           (int(k.name.split('_')[3])
                                                                            if len(k.name.split('_')) == 5 else num_chunks + 1)))

                        for idx in range(0, len(jobs_to_sort)):
                            if jobs_to_sort[idx] in fake_original_job_map:
                                fake_job = jobs_to_sort[idx]
                                jobs_to_sort[idx] = fake_original_job_map[fake_job]

                        dict_jobs[date][member] += jobs_to_sort
                        jobs_to_sort = []

                    jobs_to_sort.append(job)
                    previous_job = job

        return dict_jobs

    def _create_fake_dates_members(self, filtered_jobs_list):
        filtered_jobs_fake_date_member = []
        fake_original_job_map = dict()

        import copy
        for job in filtered_jobs_list:
            fake_job = None
            # running once and synchronize date
            if job.date is None and job.member is None:
                date = self._date_list[-1]
                member = self._member_list[-1]

                fake_job = copy.deepcopy(job)
                fake_job.name = fake_job.name.split('_', 1)[0] + "_" + self._get_date(date) + "_" \
                                + member + "_" + fake_job.name.split("_", 1)[1]
                filtered_jobs_fake_date_member.append(fake_job)
                fake_original_job_map[fake_job] = job
            # running date or synchronize member
            elif job.member is None:
                member = self._member_list[-1]
                fake_job = copy.deepcopy(job)
                fake_job.name = fake_job.name.split('_', 2)[0] + "_" + fake_job.name.split('_', 2)[
                    1] + "_" + member + "_" + fake_job.name.split("_", 2)[2]
                filtered_jobs_fake_date_member.append(fake_job)
                fake_original_job_map[fake_job] = job

            if fake_job is None:
                filtered_jobs_fake_date_member.append(job)

        return filtered_jobs_fake_date_member, fake_original_job_map

    def _get_date(self, date):
        date_format = ''
        if date.hour > 1:
            date_format = 'H'
        if date.minute > 1:
            date_format = 'M'
        str_date = date2str(date, date_format)
        return str_date

    def __len__(self):
        return self._job_list.__len__()

    def get_date_list(self):
        """
        Get inner date list

        :return: date list
        :rtype: list
        """
        return self._date_list

    def get_member_list(self):

        """
        Get inner member list

        :return: member list
        :rtype: list
        """
        return self._member_list

    def get_chunk_list(self):
        """
        Get inner chunk list

        :return: chunk list
        :rtype: list
        """
        return self._chunk_list

    def get_job_list(self):
        """
        Get inner job list

        :return: job list
        :rtype: list
        """
        return self._job_list

    def get_date_format(self):
        date_format = ''
        for date in self.get_date_list():
            if date.hour > 1:
                date_format = 'H'
            if date.minute > 1:
                date_format = 'M'
        return date_format

    def get_ordered_jobs_by_date_member(self):
        """
        Get the dictionary of jobs ordered according to wrapper's expression divided by date and member

        :return: jobs ordered divided by date and member
        :rtype: dict
        """
        return self._ordered_jobs_by_date_member

    def get_completed(self, platform=None):
        """
        Returns a list of completed jobs

        :param platform: job platform
        :type platform: HPCPlatform
        :return: completed jobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                job.status == Status.COMPLETED]
    def get_uncompleted(self, platform=None):
        """
        Returns a list of completed jobs

        :param platform: job platform
        :type platform: HPCPlatform
        :return: completed jobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                job.status != Status.COMPLETED]
    def get_submitted(self, platform=None):
        """
        Returns a list of submitted jobs

        :param platform: job platform
        :type platform: HPCPlatform
        :return: submitted jobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                job.status == Status.SUBMITTED]

    def get_running(self, platform=None):
        """
        Returns a list of jobs running

        :param platform: job platform
        :type platform: HPCPlatform
        :return: running jobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                job.status == Status.RUNNING]

    def get_queuing(self, platform=None):
        """
        Returns a list of jobs queuing

        :param platform: job platform
        :type platform: HPCPlatform
        :return: queuedjobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                job.status == Status.QUEUING]

    def get_failed(self, platform=None):
        """
        Returns a list of failed jobs

        :param platform: job platform
        :type platform: HPCPlatform
        :return: failed jobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                job.status == Status.FAILED]

    def get_unsubmitted(self, platform=None):
        """
        Returns a list of unsummited jobs

        :param platform: job platform
        :type platform: HPCPlatform
        :return: all jobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                ( job.status != Status.SUBMITTED and job.status != Status.QUEUING and job.status == Status.RUNNING and job.status == Status.COMPLETED ) ]
    def get_all(self, platform=None):
        """
        Returns a list of all jobs

        :param platform: job platform
        :type platform: HPCPlatform
        :return: all jobs
        :rtype: list
        """
        return [job for job in self._job_list]
    def get_ready(self, platform=None):
        """
        Returns a list of ready jobs

        :param platform: job platform
        :type platform: HPCPlatform
        :return: ready jobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                job.status == Status.READY]

    def get_waiting(self, platform=None):
        """
        Returns a list of jobs waiting

        :param platform: job platform
        :type platform: HPCPlatform
        :return: waiting jobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                job.status == Status.WAITING]

    def get_unknown(self, platform=None):
        """
        Returns a list of jobs on unknown state

        :param platform: job platform
        :type platform: HPCPlatform
        :return: unknown state jobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                job.status == Status.UNKNOWN]

    def get_suspended(self, platform=None):
        """
        Returns a list of jobs on unknown state

        :param platform: job platform
        :type platform: HPCPlatform
        :return: unknown state jobs
        :rtype: list
        """
        return [job for job in self._job_list if (platform is None or job.platform is platform) and
                job.status == Status.SUSPENDED]

    def get_in_queue(self, platform=None):
        """
        Returns a list of jobs in the platforms (Submitted, Running, Queuing, Unknown)

        :param platform: job platform
        :type platform: HPCPlatform
        :return: jobs in platforms
        :rtype: list
        """
        return self.get_submitted(platform) + self.get_running(platform) + self.get_queuing(
            platform) + self.get_unknown(platform)

    def get_not_in_queue(self, platform=None):
        """
        Returns a list of jobs NOT in the platforms (Ready, Waiting)

        :param platform: job platform
        :type platform: HPCPlatform
        :return: jobs not in platforms
        :rtype: list
        """
        return self.get_ready(platform) + self.get_waiting(platform)

    def get_finished(self, platform=None):
        """
        Returns a list of jobs finished (Completed, Failed)


        :param platform: job platform
        :type platform: HPCPlatform
        :return: finished jobs
        :rtype: list
        """
        return self.get_completed(platform) + self.get_failed(platform)

    def get_active(self, platform=None):
        """
        Returns a list of active jobs (In platforms, Ready)

        :param platform: job platform
        :type platform: HPCPlatform
        :return: active jobs
        :rtype: list
        """
        return self.get_in_queue(platform) + self.get_ready(platform)

    def get_job_by_name(self, name):
        """
        Returns the job that its name matches parameter name

        :parameter name: name to look for
        :type name: str
        :return: found job
        :rtype: job
        """
        for job in self._job_list:
            if job.name == name:
                return job
        Log.warning("We could not find that job {0} in the list!!!!", name)

    def get_in_queue_grouped_id(self, platform):
        jobs = self.get_in_queue(platform)
        jobs_by_id = dict()
        for job in jobs:
            if job.id not in jobs_by_id:
                jobs_by_id[job.id] = list()
            jobs_by_id[job.id].append(job)
        return jobs_by_id

    def get_in_ready_grouped_id(self, platform):
        jobs=[]
        [jobs.append(job) for job in jobs if (platform is None or job._platform.name is platform.name)]

        jobs_by_id = dict()
        for job in jobs:
            if job.id not in jobs_by_id:
                jobs_by_id[job.id] = list()
            jobs_by_id[job.id].append(job)
        return jobs_by_id

    def sort_by_name(self):
        """
        Returns a list of jobs sorted by name

        :return: jobs sorted by name
        :rtype: list
        """
        return sorted(self._job_list, key=lambda k: k.name)

    def sort_by_id(self):
        """
        Returns a list of jobs sorted by id

        :return: jobs sorted by ID
        :rtype: list
        """
        return sorted(self._job_list, key=lambda k: k.id)

    def sort_by_type(self):
        """
        Returns a list of jobs sorted by type

        :return: job sorted by type
        :rtype: list
        """
        return sorted(self._job_list, key=lambda k: k.type)

    def sort_by_status(self):
        """
        Returns a list of jobs sorted by status

        :return: job sorted by status
        :rtype: list
        """
        return sorted(self._job_list, key=lambda k: k.status)

    @staticmethod
    def load_file(filename):
        """
        Recreates an stored joblist from the pickle file

        :param filename: pickle file to load
        :type filename: str
        :return: loaded joblist object
        :rtype: JobList
        """
        if os.path.exists(filename):
            fd = open(filename, 'rw')
            return pickle.load(fd)
        else:
            Log.critical('File {0} does not exist'.format(filename))
            return list()

    def load(self):
        """
        Recreates an stored job list from the persistence

        :return: loaded job list object
        :rtype: JobList
        """
        Log.info("Loading JobList")
        return self._persistence.load(self._persistence_path, self._persistence_file)

    def save(self):
        """
        Persists the job list
        """
        self._persistence.save(self._persistence_path, self._persistence_file, self._job_list)

    def update_from_file(self, store_change=True):
        """
        Updates jobs list on the fly from and update file
        :param store_change: if True, renames the update file to avoid reloading it at the next iteration
        """
        if os.path.exists(os.path.join(self._persistence_path, self._update_file)):
            Log.info("Loading updated list: {0}".format(os.path.join(self._persistence_path, self._update_file)))
            for line in open(os.path.join(self._persistence_path, self._update_file)):
                if line.strip() == '':
                    continue
                job = self.get_job_by_name(line.split()[0])
                if job:
                    job.status = self._stat_val.retval(line.split()[1])
                    job.fail_count = 0
            now = localtime()
            output_date = strftime("%Y%m%d_%H%M", now)
            if store_change:
                move(os.path.join(self._persistence_path, self._update_file),
                     os.path.join(self._persistence_path, self._update_file +
                                  "_" + output_date))

    @property
    def parameters(self):
        """
        List of parameters common to all jobs
        :return: parameters
        :rtype: dict
        """
        return self._parameters

    @parameters.setter
    def parameters(self, value):
        self._parameters = value

    def update_list(self, as_conf,store_change=True,fromSetStatus=False):
        """
        Updates job list, resetting failed jobs and changing to READY all WAITING jobs with all parents COMPLETED

        :param as_conf: autosubmit config object
        :type as_conf: AutosubmitConfig
        :return: True if job status were modified, False otherwise
        :rtype: bool
        """
        # load updated file list
        save = False
        if self.update_from_file(store_change):
            save = store_change

        # reset jobs that has failed less than 10 times
        Log.debug('Updating FAILED jobs')
        for job in self.get_failed():
            job.inc_fail_count()
            if not hasattr(job, 'retrials') or job.retrials is None:
                retrials = as_conf.get_retrials()
            else:
                retrials = job.retrials

            if job.fail_count <= retrials:
                tmp = [parent for parent in job.parents if parent.status == Status.COMPLETED]
                if len(tmp) == len(job.parents):
                    job.status = Status.READY
                    job.packed = False
                    save = True
                    Log.debug("Resetting job: {0} status to: READY for retrial...".format(job.name))
                else:
                    job.status = Status.WAITING
                    save = True
                    job.packed = False
                    Log.debug("Resetting job: {0} status to: WAITING for parents completion...".format(job.name))

        # if waiting jobs has all parents completed change its State to READY


        for job in self.get_completed():

            if job.synchronize is not None: #and job in self.get_active():
                Log.debug('Updating SYNC jobs')
                tmp = [parent for parent in job.parents if parent.status == Status.COMPLETED]
                if len(tmp) != len(job.parents):
                    job.status = Status.WAITING
                    save = True
                    Log.debug("Resetting sync job: {0} status to: WAITING for parents completion...".format(job.name))
        Log.debug('Update finished')


        Log.debug('Updating WAITING jobs')
        for job in self.get_waiting():
            if not fromSetStatus:
                tmp = [parent for parent in job.parents if parent.status == Status.COMPLETED]
                if len(tmp) == len(job.parents):
                    job.status = Status.READY
                    save = True
                    Log.debug("Resetting job: {0} status to: READY (all parents completed)...".format(job.name))
        Log.debug('Update finished')

        return save

    def update_genealogy(self, new=True, notransitive=False):
        """
        When we have created the job list, every type of job is created.
        Update genealogy remove jobs that have no templates
        :param new: if it is a new job list or not
        :type new: bool
        """

        # Use a copy of job_list because original is modified along iterations
        for job in self._job_list[:]:
            if job.file is None or job.file == '':
                self._remove_job(job)

        # Simplifying dependencies: if a parent is already an ancestor of another parent,
        # we remove parent dependency
        if not notransitive:
            self.graph = transitive_reduction(self.graph)
            for job in self._job_list:
                children_to_remove = [child for child in job.children if child.name not in self.graph.neighbors(job.name)]
                for child in children_to_remove:
                    job.children.remove(child)
                    child.parents.remove(job)

        for job in self._job_list:
            if not job.has_parents() and new:
                job.status = Status.READY

    def check_scripts(self, as_conf):
        """
        When we have created the scripts, all parameters should have been substituted.
        %PARAMETER% handlers not allowed

        :param as_conf: experiment configuration
        :type as_conf: AutosubmitConfig
        """
        Log.info("Checking scripts...")
        out = True
        for job in self._job_list:
            if job.check.lower() == 'on_submission':
                continue
            if job.check.lower() != 'true':
                show_logs = False
                if job.section not in self.sections_checked:
                    Log.warning('Template {0} will be checked without logs'.format(job.section))
            elif job.section in self.sections_checked:
                show_logs = False
            else:
                show_logs = True

            if not job.check_script(as_conf, self.parameters,show_logs):
                out = False
                if show_logs:
                    Log.warning("Invalid parameter substitution in {0} template", job.section)

            self.sections_checked.add(job.section)
        if out:
            Log.result("Scripts OK")
        else:
            Log.warning("Scripts check failed")
            Log.user_warning("Running after failed scripts check is at your own risk!")
        return out

    def _remove_job(self, job):
        """
        Remove a job from the list

        :param job: job to remove
        :type job: Job
        """
        for child in job.children:
            for parent in job.parents:
                child.add_parent(parent)
            child.delete_parent(job)

        for parent in job.parents:
            parent.children.remove(job)

        self._job_list.remove(job)

    def rerun(self, chunk_list, notransitive=False,monitor=False):
        """
        Updates job list to rerun the jobs specified by chunk_list

        :param chunk_list: list of chunks to rerun
        :type chunk_list: str
        """
        jobs_parser = self._get_jobs_parser()

        Log.info("Adding dependencies...")
        dependencies = dict()
        for job_section in jobs_parser.sections():
            Log.debug("Reading rerun dependencies for {0} jobs".format(job_section))

            # If does not has rerun dependencies, do nothing
            if not jobs_parser.has_option(job_section, "RERUN_DEPENDENCIES"):
                continue

            dependencies_keys = jobs_parser.get(job_section, "RERUN_DEPENDENCIES").split()
            dependencies = JobList._manage_dependencies(dependencies_keys, self._dic_jobs)

        for job in self._job_list:
            job.status = Status.COMPLETED

        data = json.loads(chunk_list)
        for d in data['sds']:
            date = parse_date(d['sd'])
            Log.debug("Date: {0}", date)
            for m in d['ms']:
                member = m['m']
                Log.debug("Member: " + member)
                previous_chunk = 0
                for c in m['cs']:
                    Log.debug("Chunk: " + c)
                    chunk = int(c)
                    for job in [i for i in self._job_list if i.date == date and i.member == member and  (i.chunk == chunk ) ]:

                        if not job.rerun_only or chunk != previous_chunk + 1:
                            job.status = Status.WAITING
                            Log.debug("Job: " + job.name)

                        job_section = job.section
                        if job_section not in dependencies:
                            continue

                        for key in dependencies_keys:
                            skip, (current_chunk, current_member, current_date) = JobList._calculate_dependency_metadata(chunk, member, date,
                                                                                                 dependencies[key])
                            if skip:
                                continue

                            section_name = dependencies[key].section
                            for parent in self._dic_jobs.get_jobs(section_name, current_date, current_member,
                                                                  current_chunk):
                                parent.status = Status.WAITING
                                Log.debug("Parent: " + parent.name)


        for job in [j for j in self._job_list if j.status == Status.COMPLETED]:
            if job.synchronize is None:
                self._remove_job(job)

        self.update_genealogy(notransitive=notransitive)
        for job in [j for j in self._job_list if j.synchronize !=None]:
            if job.status == Status.COMPLETED:
                job.status = Status.WAITING
            else:
                self._remove_job(job)

    def _get_jobs_parser(self):
        jobs_parser = self._parser_factory.create_parser()
        jobs_parser.optionxform = str
        jobs_parser.read(
            os.path.join(self._config.LOCAL_ROOT_DIR, self._expid, 'conf', "jobs_" + self._expid + ".conf"))
        return jobs_parser

    def remove_rerun_only_jobs(self, notransitive=False):
        """
        Removes all jobs to be run only in reruns
        """
        flag = False
        for job in set(self._job_list):
            if job.rerun_only:
                self._remove_job(job)
                flag = True

        if flag:
            self.update_genealogy(notransitive=notransitive)
        del self._dic_jobs



#!/usr/bin/env python

# Copyright 2017 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

from autosubmit.job.job import Job
from bscearth.utils.date import date2str
from autosubmit.job.job_common import Status, Type


class DicJobs:
    """
    Class to create jobs from conf file and to find jobs by start date, member and chunk

    :param jobs_list: jobs list to use
    :type job_list: JobList
    :param parser: jobs conf file parser
    :type parser: SafeConfigParser
    :param date_list: start dates
    :type date_list: list
    :param member_list: member
    :type member_list: list
    :param chunk_list: chunks
    :type chunk_list: list
    :param date_format: option to format dates
    :type date_format: str
    :param default_retrials: default retrials for ech job
    :type default_retrials: int

    """

    def __init__(self, jobs_list, parser, date_list, member_list, chunk_list, date_format, default_retrials):
        self._date_list = date_list
        self._jobs_list = jobs_list
        self._member_list = member_list
        self._chunk_list = chunk_list
        self._parser = parser
        self._date_format = date_format
        self.default_retrials = default_retrials
        self._dic = dict()

    def read_section(self, section, priority, default_job_type, jobs_data=dict()):
        """
        Read a section from jobs conf and creates all jobs for it

        :param default_job_type: default type for jobs
        :type default_job_type: str
        :param jobs_data: dictionary containing the plain data from jobs
        :type jobs_data: dict
        :param section: section to read
        :type section: str
        :param priority: priority for the jobs
        :type priority: int
        """
        running = 'once'
        if self._parser.has_option(section, 'RUNNING'):
            running = self._parser.get(section, 'RUNNING').lower()
        frequency = int(self.get_option(section, "FREQUENCY", 1))
        if running == 'once':
            self._create_jobs_once(section, priority, default_job_type, jobs_data)
        elif running == 'date':
            self._create_jobs_startdate(section, priority, frequency, default_job_type, jobs_data)
        elif running == 'member':
            self._create_jobs_member(section, priority, frequency, default_job_type, jobs_data)
        elif running == 'chunk':
            synchronize = self.get_option(section, "SYNCHRONIZE", None)
            delay = int(self.get_option(section, "DELAY", -1))
            splits = int(self.get_option(section, "SPLITS", 0))
            self._create_jobs_chunk(section, priority, frequency, default_job_type, synchronize, delay, splits, jobs_data)

    def _create_jobs_once(self, section, priority, default_job_type, jobs_data=dict()):
        """
        Create jobs to be run once

        :param section: section to read
        :type section: str
        :param priority: priority for the jobs
        :type priority: int
        """
        self._dic[section] = self.build_job(section, priority, None, None, None, default_job_type, jobs_data)
        self._jobs_list.graph.add_node(self._dic[section].name)

    def _create_jobs_startdate(self, section, priority, frequency, default_job_type, jobs_data=dict()):
        """
        Create jobs to be run once per start date

        :param section: section to read
        :type section: str
        :param priority: priority for the jobs
        :type priority: int
        :param frequency: if greater than 1, only creates one job each frequency startdates. Always creates one job
                          for the last
        :type frequency: int
        """
        self._dic[section] = dict()
        count = 0
        for date in self._date_list:
            count += 1
            if count % frequency == 0 or count == len(self._date_list):
                self._dic[section][date] = self.build_job(section, priority, date, None, None, default_job_type,
                                                          jobs_data)
                self._jobs_list.graph.add_node(self._dic[section][date].name)

    def _create_jobs_member(self, section, priority, frequency, default_job_type, jobs_data=dict()):
        """
        Create jobs to be run once per member

        :param section: section to read
        :type section: str
        :param priority: priority for the jobs
        :type priority: int
        :param frequency: if greater than 1, only creates one job each frequency members. Always creates one job
                          for the last
        :type frequency: int
        """
        self._dic[section] = dict()
        for date in self._date_list:
            self._dic[section][date] = dict()
            count = 0
            for member in self._member_list:
                count += 1
                if count % frequency == 0 or count == len(self._member_list):
                    self._dic[section][date][member] = self.build_job(section, priority, date, member, None,
                                                                      default_job_type, jobs_data)
                    self._jobs_list.graph.add_node(self._dic[section][date][member].name)

    '''
        Maybe a good choice could be split this function or ascend the
        conditional decision to the parent which makes the call
    '''

    def _create_jobs_chunk(self, section, priority, frequency, default_job_type, synchronize=None, delay=0, splits=0, jobs_data=dict()):
        """
        Create jobs to be run once per chunk

        :param synchronize:
        :param section: section to read
        :type section: str
        :param priority: priority for the jobs
        :type priority: int
        :param frequency: if greater than 1, only creates one job each frequency chunks. Always creates one job
                          for the last
        :type frequency: int
        :param delay: if this parameter is set, the job is only created for the chunks greater than the delay
        :type delay: int
        """
        # Temporally creation for unified jobs in case of synchronize

        if synchronize is not None:
            tmp_dic = dict()
            count = 0
            for chunk in self._chunk_list:
                count += 1
                if delay == -1 or delay < chunk:
                    if count % frequency == 0 or count == len(self._chunk_list):
                        if splits > 1:
                            if synchronize == 'date':
                                tmp_dic[chunk] = []
                                self._create_jobs_split(splits, section, None, None, chunk, priority,
                                                   default_job_type, jobs_data, tmp_dic[chunk])
                            elif synchronize == 'member':
                                tmp_dic[chunk] = dict()
                                for date in self._date_list:
                                    tmp_dic[chunk][date] = []
                                    self._create_jobs_split(splits, section, date, None, chunk, priority,
                                                            default_job_type, jobs_data, tmp_dic[chunk][date])

                        else:
                            if synchronize == 'date':
                                tmp_dic[chunk] = self.build_job(section, priority, None, None,
                                                                chunk, default_job_type, jobs_data)
                            elif synchronize == 'member':
                                tmp_dic[chunk] = dict()
                                for date in self._date_list:
                                    tmp_dic[chunk][date] = self.build_job(section, priority, date, None,
                                                                      chunk, default_job_type, jobs_data)
        # Real dic jobs assignment/creation
        self._dic[section] = dict()
        for date in self._date_list:
            self._dic[section][date] = dict()
            for member in self._member_list:
                self._dic[section][date][member] = dict()
                count = 0
                for chunk in self._chunk_list:
                    count += 1
                    if delay == -1 or delay < chunk:
                        if count % frequency == 0 or count == len(self._chunk_list):
                            if synchronize == 'date':
                                self._dic[section][date][member][chunk] = tmp_dic[chunk]
                            elif synchronize == 'member':
                                self._dic[section][date][member][chunk] = tmp_dic[chunk][date]

                            if splits > 1 and synchronize is None:
                                self._dic[section][date][member][chunk] = []
                                self._create_jobs_split(splits, section, date, member, chunk, priority, default_job_type, jobs_data, self._dic[section][date][member][chunk])
                            elif synchronize is None:
                                self._dic[section][date][member][chunk] = self.build_job(section, priority, date, member,
                                                                                             chunk, default_job_type, jobs_data)
                                self._jobs_list.graph.add_node(self._dic[section][date][member][chunk].name)

    def _create_jobs_split(self, splits, section, date, member, chunk, priority, default_job_type, jobs_data, dict):
        total_jobs = 1
        while total_jobs <= splits:
            job = self.build_job(section, priority, date, member, chunk, default_job_type, jobs_data, total_jobs)
            dict.append(job)
            self._jobs_list.graph.add_node(job.name)
            total_jobs += 1

    def get_jobs(self, section, date=None, member=None, chunk=None):
        """
        Return all the jobs matching section, date, member and chunk provided. If any parameter is none, returns all
        the jobs without checking that parameter value. If a job has one parameter to None, is returned if all the
        others match parameters passed

        :param section: section to return
        :type section: str
        :param date: stardate to return
        :type date: str
        :param member: member to return
        :type member: str
        :param chunk: chunk to return
        :type chunk: int
        :return: jobs matching parameters passed
        :rtype: list
        """
        jobs = list()

        if section not in self._dic:
            return jobs

        dic = self._dic[section]
        if type(dic) is not dict:
            jobs.append(dic)
        else:
            if date is not None:
                self._get_date(jobs, dic, date, member, chunk)
            else:
                for d in self._date_list:
                    self._get_date(jobs, dic, d, member, chunk)
        return jobs

    def _get_date(self, jobs, dic, date, member, chunk):
        if date not in dic:
            return jobs
        dic = dic[date]
        if type(dic) is not dict:
            jobs.append(dic)
        else:
            if member is not None:
                self._get_member(jobs, dic, member, chunk)
            else:
                for m in self._member_list:
                    self._get_member(jobs, dic, m, chunk)

        return jobs

    def _get_member(self, jobs, dic, member, chunk):
        if member not in dic:
            return jobs
        dic = dic[member]
        if type(dic) is not dict:
            jobs.append(dic)
        else:
            if chunk is not None:
                if chunk in dic:
                    jobs.append(dic[chunk])
            else:
                for c in self._chunk_list:
                    if c not in dic:
                        continue
                    jobs.append(dic[c])
        return jobs

    def build_job(self, section, priority, date, member, chunk, default_job_type, jobs_data=dict(), split=-1):
        name = self._jobs_list.expid
        if date is not None:
            name += "_" + date2str(date, self._date_format)
        if member is not None:
            name += "_" + member
        if chunk is not None:
            name += "_{0}".format(chunk)
        if split > -1:
            name += "_{0}".format(split)
        name += "_" + section
        if name in jobs_data:
            job = Job(name, jobs_data[name][1], jobs_data[name][2], priority)
            job.local_logs = (jobs_data[name][8], jobs_data[name][9])
            job.remote_logs = (jobs_data[name][10], jobs_data[name][11])

        else:
            job = Job(name, 0, Status.WAITING, priority)


        job.section = section
        job.date = date
        job.member = member
        job.chunk = chunk
        job.date_format = self._date_format
        if split > -1:
            job.split = split

        job.frequency = int(self.get_option(section, "FREQUENCY", 1))
        job.delay = int(self.get_option(section, "DELAY", -1))
        job.wait = self.get_option(section, "WAIT", 'true').lower() == 'true'
        job.rerun_only = self.get_option(section, "RERUN_ONLY", 'false').lower() == 'true'

        job_type = self.get_option(section, "TYPE", default_job_type).lower()
        if job_type == 'bash':
            job.type = Type.BASH
        elif job_type == 'python':
            job.type = Type.PYTHON
        elif job_type == 'r':
            job.type = Type.R

        job.platform_name = self.get_option(section, "PLATFORM", None)
        if job.platform_name is not None:
            job.platform_name = job.platform_name
        job.file = self.get_option(section, "FILE", None)
        job.queue = self.get_option(section, "QUEUE", None)
        job.check = self.get_option(section, "CHECK", 'True').lower()
        job.processors = str(self.get_option(section, "PROCESSORS", 1))
        job.threads = str(self.get_option(section, "THREADS", 1))
        job.tasks = str(self.get_option(section, "TASKS", '0'))
        job.memory = self.get_option(section, "MEMORY", '')
        job.memory_per_task = self.get_option(section, "MEMORY_PER_TASK", '')
        job.wallclock = self.get_option(section, "WALLCLOCK", '')


        job.retrials = int(self.get_option(section, 'RETRIALS', -1))

        if job.retrials == -1:
            job.retrials = None
        job.notify_on = [x.upper() for x in self.get_option(section, "NOTIFY_ON", '').split(' ')]
        job.synchronize = self.get_option(section, "SYNCHRONIZE", None)

        self._jobs_list.get_job_list().append(job)

        return job

    def get_option(self, section, option, default):
        """
        Returns value for a given option

        :param section: section name
        :type section: str
        :param option: option to return
        :type option: str
        :param default: value to return if not defined in configuration file
        :type default: object
        """
        if self._parser.has_option(section, option):
            return self._parser.get(section, option)
        else:
            return default



from tests_log import Log
from tests_utils import check_cmd, next_experiment_id, copy_experiment_conf_files, create_database, clean_database
from tests_commands import *
from threading import Thread
from time import sleep
import argparse
try:
    # noinspection PyCompatibility
    from configparser import SafeConfigParser
except ImportError:
    # noinspection PyCompatibility
    from ConfigParser import SafeConfigParser

# Configuration file where the regression tests are defined with INI style
tests_parser_file = 'tests.conf'

# Path where the temporal files (db, experiments, etc) fill be saved
db_path = './db'

# Initial experiment id
initial_experiment_id = 'a000'


def run_test_case(experiment_id, name, hpc_arch, description, src_path, retrials=2):
    if not check_cmd(generate_experiment_cmd(hpc_arch, description)):
        Log.error('Error while generating the experiment {0}({1})', name, experiment_id)
        return False
    copy_experiment_conf_files(db_path, src_path, experiment_id)
    sleep(5)  # Avoiding synchronization problems while copying

    run_ok = False
    for attempt in range(retrials):
        if not check_cmd(create_experiment_cmd(experiment_id)):
            Log.warning('Error while creating the experiment {0}({1})', name, experiment_id)
            continue

        if not check_cmd(check_experiment_cmd(experiment_id)):
            Log.warning('Error while checking the experiment {0}({1})', name, experiment_id)
            continue

        if not check_cmd(monitor_experiment_cmd(experiment_id)):
            Log.warning('Error while monitoring the experiment {0}({1})', name, experiment_id)
            continue

        if not check_cmd(refresh_experiment_cmd(experiment_id)):
            Log.warning('Error while refreshing the experiment {0}({1})', name, experiment_id)
            continue

        if not check_cmd(run_experiment_cmd(experiment_id)):
            Log.warning('Error while running the experiment {0}({1})', name, experiment_id)
            continue

        if not check_cmd(monitor_experiment_cmd(experiment_id)):
            Log.warning('Error while monitoring the experiment {0}({1})', name, experiment_id)
            continue

        if not check_cmd(stats_experiment_cmd(experiment_id)):
            Log.warning('Error while getting stats of the experiment {0}({1})', name, experiment_id)
            continue

        if not check_cmd(recovery_experiment_cmd(experiment_id)):
            Log.warning('Error while recovering the experiment {0}({1})', name, experiment_id)
            continue

        run_ok = True
        break
    if run_ok:
        Log.result('[OK] Test {0}({1}) has been passed successfully', name, experiment_id)
    else:
        Log.error('[KO] Test {0}({1}) has not been passed successfully', name, experiment_id)


def run(current_experiment_id, only_list=None, exclude_list=None, max_threads=5):
    # Local variables for testing
    test_threads = []

    # Building tests parser
    tests_parser = SafeConfigParser()
    tests_parser.optionxform = str
    tests_parser.read(tests_parser_file)

    # Resetting the database
    clean_database(db_path)
    create_database()

    # Main loop to run all the tests
    for section in tests_parser.sections():
        # Skipping filtered experiments
        if only_list is not None and section not in only_list:
            Log.warning('Test {0} has been skipped', section)
            continue

        if exclude_list is not None and section in exclude_list:
            Log.warning('Test {0} has been skipped', section)
            continue

        # Prevents too many concurrent threads
        if len(test_threads) >= max_threads:
            test_threads.pop(0).join()

        # Getting test settings
        description, hpc_arch, src_path = get_test_settings(section, tests_parser)

        # Running the test as a new thread
        test_threads.append(create_test_thread(current_experiment_id, section, description, hpc_arch, src_path))

        # Updating current experiment id
        current_experiment_id = next_experiment_id(current_experiment_id)

        # Avoiding synchronization problems
        sleep(10)

    # Loop to wait the end of all the running tests
    for test_thread in test_threads:
        test_thread.join()


def create_test_thread(current_experiment_id, name, description, hpc_arch, src_path):
    thr = Thread(target=run_test_case, args=(current_experiment_id, name, hpc_arch, description, src_path), kwargs={})
    thr.start()
    return thr


def get_test_settings(section, tests_parser):
    hpc_arch = tests_parser.get(section, 'HPCARCH')
    description = tests_parser.get(section, 'DESCRIPTION')
    src_path = tests_parser.get(section, 'SRC_PATH')
    return description, hpc_arch, src_path


def create_args_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--only", type=str,
                        help="List of experiments to be run, test names separated by white spaces")
    parser.add_argument("--exclude", type=str,
                        help="List of experiments to be avoided, test names separated by white spaces")
    return parser


if __name__ == "__main__":
    args_parser = create_args_parser()

    only = None if args_parser.parse_args().only is None else args_parser.parse_args().only.split()
    exclude = None if args_parser.parse_args().exclude is None else args_parser.parse_args().exclude.split()

    run(initial_experiment_id, only, exclude)
